import unittest

__author__ = "diarmuid"
import sys

sys.path.append("..")
import unittest
import AcraNetwork.Pcap as pcap
import AcraNetwork.SimpleEthernet as SimpleEthernet
import AcraNetwork.IRIG106.Chapter11 as ch10
import AcraNetwork.IRIG106.Chapter10.Chapter10UDP as ch10udp
import AcraNetwork.IRIG106.Chapter11.ARINC429 as ch10arinc
import AcraNetwork.IRIG106.Chapter11.UART as ch10uart
import AcraNetwork.IRIG106.Chapter11.TimeDataFormat as ch10time
import AcraNetwork.IRIG106.Chapter11.ARINC429 as ch10arinc
import AcraNetwork.IRIG106.Chapter11.MILSTD1553 as ch10mil
import AcraNetwork.IRIG106.Chapter11.PCM as ch10pcm
import AcraNetwork.IRIG106.Chapter11.Video as ch10video
import AcraNetwork.IRIG106
from AcraNetwork.IRIG106.Chapter11 import (
    DATA_TYPE_TIMEFMT_1,
    DATA_TYPE_TIMEFMT_2,
    DATA_TYPE_PCM_DATA_FMT1,
    TS_CH4,
    TS_IEEE1558,
    RTCTime,
    PTPTime,
)
import struct
import os
import os.path
import random
import copy
import datetime
import tempfile
import logging
import csv
import AcraNetwork.MPEGTS as ampeg
import base64

THIS_DIR = os.path.dirname(os.path.abspath(__file__))
TMP_DIR = tempfile.gettempdir()


logging.basicConfig(level=logging.INFO)
# logging.info(f"Temp folder={TMP_DIR}")


def getEthernetPacket(data: bytes = b""):
    e = SimpleEthernet.Ethernet()
    e.srcmac = 0x001122334455
    e.dstmac = 0x998877665544
    e.type = SimpleEthernet.Ethernet.TYPE_IP
    i = SimpleEthernet.IP()
    i.dstip = "235.0.0.1"
    i.srcip = "192.168.1.1"
    i.protocol = SimpleEthernet.IP.PROTOCOLS["UDP"]
    u = SimpleEthernet.UDP()
    u.dstport = 51000
    u.srcport = 51000
    u.payload = data
    i.payload = u.pack()
    e.payload = i.pack()
    return e.pack()


def wrap_in_udp_and_pcap(mybuffer, pcapf, mode="w"):
    pcapw = pcap.Pcap(pcapf, mode=mode)

    rec = pcap.PcapRecord()
    rec.payload = getEthernetPacket(mybuffer)
    pcapw.write(rec)
    pcapw.close()
    return True


arinc_packet = """ARINCPayload: MessageCount=11
  ARINCData: GapTime=0 FormatError=False ParityError=False BusSpeed=0 Bus=23
  ARINCData: GapTime=4000 FormatError=False ParityError=False BusSpeed=0 Bus=23
  ARINCData: GapTime=3900 FormatError=False ParityError=False BusSpeed=0 Bus=23
  ARINCData: GapTime=3600 FormatError=False ParityError=False BusSpeed=0 Bus=23
  ARINCData: GapTime=4100 FormatError=False ParityError=False BusSpeed=0 Bus=23
  ARINCData: GapTime=4000 FormatError=False ParityError=False BusSpeed=0 Bus=23
  ARINCData: GapTime=3700 FormatError=False ParityError=False BusSpeed=0 Bus=23
  ARINCData: GapTime=3600 FormatError=False ParityError=False BusSpeed=0 Bus=23
  ARINCData: GapTime=4000 FormatError=False ParityError=False BusSpeed=0 Bus=23
  ARINCData: GapTime=3600 FormatError=False ParityError=False BusSpeed=0 Bus=23
  ARINCData: GapTime=3900 FormatError=False ParityError=False BusSpeed=0 Bus=23
"""

arinc_updated = base64.b64decode(
    """AQBeAAABABEiM0RVCABFAABYAABAABQRnunAqBwB6wAAAhFcGhcARAAAAQsAACXrAQo4AAAAFAAAAAYAxDg0OgAAAABwaPi7FgAAAAAAAADJAQIAAAAAACABlAEAQIBwIAGWBgAg7QSKXA=="""
)


def get_ch10(len=4):
    c = ch10.Chapter11()
    c.channelID = 1
    c.datatypeversion = 2
    c.sequence = 3
    c.packetflag = 0  # No secondary
    c.datatype = 4
    c.relativetimecounter = 100
    c.has_secondary_header = False
    c.payload = os.urandom(len)
    return c


class CH10UDPTest(unittest.TestCase):
    def setUp(self):
        self.full = ch10udp.Chapter10UDP()
        self.full.type = ch10udp.Chapter10UDP.TYPE_FULL
        self.full.sequence = 1
        self.ch10 = ch10.Chapter11()
        self.ch10.channelID = 1
        self.ch10.datatypeversion = 2
        self.ch10.sequence = 3
        self.ch10.packetflag = 0  # No secondary
        self.ch10.datatype = 4
        self.ch10.relativetimecounter = 100
        self.full.payload = self.ch10.pack()

        self.seg = ch10udp.Chapter10UDP()
        self.seg.sequence = 2
        self.seg.type = ch10udp.Chapter10UDP.TYPE_SEG
        self.seg.channelID = 0x232
        self.seg.channelsequence = 100
        self.seg.segmentoffset = 3
        self.ch10seg = ch10.Chapter11()
        self.ch10seg.channelID = 1
        self.ch10seg.datatypeversion = 2
        self.ch10seg.sequence = 3
        self.ch10seg.packetflag = 0  # No secondary
        self.ch10seg.datatype = 4
        self.ch10seg.relativetimecounter = 100
        self.seg.payload = self.ch10seg.pack()

        self.c = ch10.Chapter11()
        self.c.channelID = 1
        self.c.datatypeversion = 2
        self.c.sequence = 3
        self.c.packetflag = 0  # No secondary
        self.c.datatype = 4
        self.c.relativetimecounter = 100

    def test_ch10_to_pcap(self):
        # self.full._payload = struct.pack(">II", 33, 44)
        full_payload = getEthernetPacket(self.full.pack())
        self.pcapw = pcap.Pcap(TMP_DIR + "/test_ch10.pcap", mode="w")

        self.rec = pcap.PcapRecord()
        self.rec.payload = full_payload
        self.assertIsNone(self.pcapw.write(self.rec))
        # self.seg._payload = struct.pack(">II", 55, 66)
        seg_payload = getEthernetPacket(self.seg.pack())
        self.rec.payload = seg_payload
        self.assertIsNone(self.pcapw.write(self.rec))
        self.pcapw.close()

    def test_ch10_eq(self):
        full = ch10udp.Chapter10UDP()
        full.type = ch10udp.Chapter10UDP.TYPE_FULL
        full.sequence = 1
        full.payload = self.full.payload
        # full._payload = struct.pack(">II", 33, 44)

        # self.full._payload = struct.pack(">II", 33, 44)

        self.assertTrue(full == self.full)

    def test_unpack(self):
        # self.full._payload = struct.pack(">II", 33, 44)
        # self.seg._payload = struct.pack(">II", 33, 44)

        full_unpack = ch10udp.Chapter10UDP()
        self.assertTrue(full_unpack.unpack(self.full.pack()))
        # print(len(self.full.payload))
        # print(len(full_unpack.payload))

        self.assertTrue(full_unpack == self.full)
        self.assertFalse(full_unpack == self.seg)

    def test_ch10_eqseg(self):
        seg = ch10udp.Chapter10UDP()
        seg.type = ch10udp.Chapter10UDP.TYPE_SEG
        seg.sequence = 2
        seg.channelID = 0x232
        seg.channelsequence = 100
        seg.segmentoffset = 3
        seg.payload = self.ch10seg.pack()
        # seg._payload = struct.pack(">II", 33, 44)

        # self.seg._payload = struct.pack(">II", 33, 44)

        self.assertTrue(seg == self.seg)

    def test_ch10pay_to_pcap(self):
        self.c.payload = struct.pack(">II", 33, 44)
        # self.full._payload = self.c.pack()
        full_payload = getEthernetPacket(self.full.pack())
        self.pcapw = pcap.Pcap(TMP_DIR + "/test_ch10_2.pcap", mode="w")

        self.rec = pcap.PcapRecord()
        self.rec.payload = full_payload
        self.assertIsNone(self.pcapw.write(self.rec))
        self.pcapw.close()

    def test_ch10_comp(self):
        ref = ch10.Chapter11()
        # self.c._payload = struct.pack(">II", 33, 44)
        pay = self.c.pack()
        ref.unpack(pay)
        self.assertTrue(ref == self.c)

    def test_ch10_sec_hdr(self):
        c = ch10udp.Chapter10UDP()
        c.type = ch10udp.Chapter10UDP.TYPE_FULL
        c.sequence = 1
        d = ch10.Chapter11()
        d.channelID = 1
        d.datatypeversion = 2
        d.sequence = 3
        d.packetflag = ch10.Chapter11.PKT_FLAG_SECONDARY
        d.datatype = 4
        d.ts_source = ch10.TS_IEEE1558
        d.ptptime = PTPTime(101, int(200e6))
        d.relativetimecounter = 0x0
        d.payload = struct.pack(">QQ", 33, 44)
        full_payload = getEthernetPacket(c.pack())
        self.pcapw = pcap.Pcap(TMP_DIR + "/test_ch10_3.pcap", mode="w")

        self.rec = pcap.PcapRecord()
        self.rec.payload = full_payload
        self.assertIsNone(self.pcapw.write(self.rec))
        self.pcapw.close()

    def test_ch10_format2(self):
        self.pcapw = pcap.Pcap(TMP_DIR + "/test_ch10_format2.pcap", mode="w")

        self.rec = pcap.PcapRecord()
        c = ch10udp.Chapter10UDP()
        c.format = 2
        c.type = ch10udp.Chapter10UDP.TYPE_FULL
        c.sequence = 10
        c.segmentoffset = 5000
        c.channelID = 20
        d = get_ch10(16)
        self.assertTrue(c.pack())

        full_payload = getEthernetPacket(c.pack())
        self.rec.payload = full_payload
        self.assertIsNone(self.pcapw.write(self.rec))
        self.pcapw.close()

        c2 = ch10udp.Chapter10UDP()
        c2.unpack(c.pack())
        self.assertTrue(c2 == c)

    def test_ch10_format3(self):
        pcapw = pcap.Pcap(TMP_DIR + "/test_ch10_format3.pcap", mode="w")

        rec = pcap.PcapRecord()
        for len in range(5):
            c = ch10udp.Chapter10UDP()

            c.format = 3
            c.sourceid_len = len
            if len > 0:
                c.sourceid = 0x4
            c.offset_pkt_start = 4
            c.sequence = 0x15
            c.payload = get_ch10(16).pack()
            self.assertTrue(c.pack())
            # print(repr(c))
            c2 = ch10udp.Chapter10UDP()
            c2.unpack(c.pack())
            rec.payload = getEthernetPacket(c.pack())
            pcapw.write(rec)
            rec.payload = getEthernetPacket(c2.pack())
            pcapw.write(rec)
            self.assertTrue(c2 == c)
            for packet in [c, c2]:
                full_payload = getEthernetPacket(packet.pack())
                rec.payload = full_payload
                self.assertIsNone(pcapw.write(rec))
        pcapw.close()

    def test_arinc_payload(self):
        p = pcap.Pcap(os.path.join(THIS_DIR, "ch10_arinc.pcap"))
        mypcaprecord = p[0]
        p.close()
        e = SimpleEthernet.Ethernet()
        e.unpack(mypcaprecord.packet)
        ip = SimpleEthernet.IP()
        ip.unpack(e.payload)
        u = SimpleEthernet.UDP()
        u.unpack(ip.payload)
        # Now I have a _payload that will be an inetx packet
        c = ch10udp.Chapter10UDP()
        self.assertTrue(c.unpack(u.payload))
        d = ch10.Chapter11()
        d.unpack(c.payload)

        # Check the Ch10 packets
        self.assertEqual(d.syncpattern, 0xEB25)
        self.assertEqual(d.channelID, 0x000080D6)
        self.assertEqual(d.datatypeversion, 0x44)
        # nanoseconds and seconds
        self.assertEqual(d.ptptime.nanoseconds, 1237463)
        self.assertEqual(d.ptptime.seconds, 0)

        self.assertEqual(repr(c), "CH10 UDP Full Packet: Format=1 Sequence=0")
        arinc_p = ch10arinc.ARINC429DataPacket()
        self.assertTrue(arinc_p.unpack(d.payload))
        self.assertEqual(len(arinc_p.arincwords), 11)
        self.assertEqual(arinc_p.arincwords[0].payload, struct.pack(">I", 0x4BD91E2E))

        arinc2 = ch10arinc.ARINC429DataPacket()
        self.assertTrue(arinc2.unpack(arinc_p.pack()))
        self.assertTrue(arinc_p == arinc2)

        # self.assertEqual(repr(arinc2), arinc_packet)
        for idx, aw in enumerate(arinc2):
            if idx == 2:
                self.assertEqual(
                    repr(aw), "ARINCData: GapTime=15375 FormatError=False ParityError=False BusSpeed=0 Bus=23"
                )

    def test_arinc2(self):
        dp = ch10arinc.ARINC429DataPacket()
        dp.unpack(arinc_updated[0x52:])
        print(repr(dp))
        for adw in dp:
            self.assertEqual(adw.bus_speed, 1)

    @unittest.skip("No trying to guess format1")
    def test_format1_looks_like_2(self):
        p = pcap.Pcap(os.path.join(THIS_DIR, "ch10_format_1_looks_like2.pcap"))
        for r in p:
            c = ch10udp.Chapter10UDP()
            c.unpack(r.payload[0x2A:])
            self.assertEqual(c.format, 1)
        p.close()


uart_pkt = """UARTPayload: UARTDataWordCount=1
  UARTDataWord: Time=PTP: 14:21:43 20-May 1976 nanosec=10 ParityError=False DataLen=508 SubChannel=8191 Endianness=<Endianness.BIG: 0>
"""


class Ch10UARTTest(unittest.TestCase):
    def setUp(self):
        self.full = ch10udp.Chapter10UDP()
        self.full.type = ch10udp.Chapter10UDP.TYPE_FULL
        self.full.sequence = 10
        self.chfull = ch10.Chapter11()
        self.chfull.channelID = 23
        self.chfull.datatypeversion = 2
        self.chfull.sequence = 3
        self.chfull.packetflag = 0xC4  # Secondary time + PTP
        self.chfull.datatype = 0x50
        self.chfull.relativetimecounter = 100
        self.chfull.ptptime = PTPTime(22, 250000)
        self.full.payload = self.chfull.pack()

    def test_basic_uart(self):
        udp = ch10uart.UARTDataPacket(TS_IEEE1558)
        for i in range(5):
            udw = ch10uart.UARTDataWord()
            udw.ipts.seconds = 22 * i
            udw.ipts.nanoseconds = 250000 + i
            udw.subchannel = 52 + 2 * i
            udw.parity_error = random.choice([True, False])
            udw.payload = os.urandom(random.randint(15, 220))
            # udw.payload = os.urandom(16)

            self.assertIsNone(udp.append(udw))

        self.chfull.payload = udp.pack()
        self.assertTrue(wrap_in_udp_and_pcap(self.full.pack(), "ch10_uart.pcap"))

        # Test unpack and comparsion
        dummy_ch10 = ch10udp.Chapter10UDP()
        dummy_ch10.unpack(self.full.pack())
        self.assertTrue(dummy_ch10 == self.full)
        self.assertEqual(
            repr(udp[0]),
            "UARTDataWord: Time=RTC: count=0 ParityError={} DataLen={} "
            "SubChannel=52 Endianness=<Endianness.BIG: 0>".format(udp[0].parity_error, udp[0].datalength),
        )

    def test_uart_unpack(self):
        p = pcap.Pcap(os.path.join(THIS_DIR, "ch10_uart2.pcap"))
        mypcaprecord = p[0]
        ch10udppkt = ch10udp.Chapter10UDP()
        ch10udppkt.unpack(mypcaprecord.payload[0x2A:-4])  # FCS
        ch10pkt = ch10.Chapter11()
        ch10pkt.unpack(ch10udppkt.payload)
        self.assertEqual(ch10pkt.packetflag, 0xC4)
        uart = ch10uart.UARTDataPacket(TS_IEEE1558)
        uart.unpack(ch10pkt.payload)
        self.assertEqual(len(uart), 1)
        (b1, b2) = struct.unpack_from("<BB", uart[0].payload)
        self.assertEqual(b1, 0xD5)
        p.close()
        for idx, uartdw in enumerate(uart):
            if idx == 0:
                dw2 = copy.copy(uartdw)
                self.assertEqual(dw2, uartdw)
                # print("{}:{}:{}".format(idx,s_idx,0))

        uart2 = copy.copy(uart)
        self.assertEqual(uart, uart2)
        self.assertEqual(repr(uart2), uart_pkt)

    @unittest.skip("missing pcap")
    def test_uart_le_unpack(self):
        p = pcap.Pcap(os.path.join(THIS_DIR, "ch10_uart_le.pcap"))
        mypcaprecord = p[0]
        ch10pkt = ch10.Chapter11()
        ch10pkt.unpack(mypcaprecord.payload[0x2E:-4])
        uart = ch10uart.UARTDataPacket(TS_IEEE1558, ch10uart.Endianness.LITTLE)
        uart.unpack(ch10pkt.payload)
        self.assertEqual(len(uart), 3)
        for idx, dw in enumerate(uart):
            if idx == 0:
                self.assertEqual(dw.datalength, 365)
                self.assertEqual(len(dw.payload), 365)
                (last_byte,) = struct.unpack_from(">B", dw.payload, 365 - 1)
                self.assertEqual(last_byte, 0x4B)

    @unittest.skip("")
    def test_fmt2_unpack(self):
        p = pcap.Pcap(os.path.join(THIS_DIR, "ch10_fmt_1.pcap"))
        mypcaprecord = p[0]
        ch10pkt = ch10udp.Chapter10UDP()
        ch10pkt.unpack(mypcaprecord.payload[0x2A:-4])  # FCS
        self.assertEqual(ch10pkt.format, 1)
        # print(repr(ch10pkt.chapter10))


class Time_Test(unittest.TestCase):
    def test_playground(self):
        ms = 989
        self.assertEqual(0x98, ch10time.double_digits_to_bcd(ms / 10))
        self.assertEqual(0x98, ch10time.double_digits_to_bcd(ms / 10))

        # 05/20/19 16:41:24'
        current_time = datetime.datetime.fromtimestamp(1558366884)
        self.assertEqual(0x41, ch10time.double_digits_to_bcd(current_time.minute))
        self.assertEqual(0x24, ch10time.double_digits_to_bcd(current_time.second))
        # self.assertEqual(0x16, ch10.double_digits_to_bcd(current_time.hour))

        self.assertEqual(99, ch10time.bcd_to_int(0x99))
        self.assertEqual(9999, ch10time.bcd_to_int(0x9999))

    def test_time_pkt(self):
        t = ch10time.TimeDataFormat1()
        t.ptptime = PTPTime(1558366884, 980000000)
        # t.milliseconds = 980
        # t.datetime = datetime.datetime.fromtimestamp(1558366884)

        # print(repr(t))
        t2 = ch10time.TimeDataFormat1()
        t2.unpack(t.pack())
        self.assertTrue(t == t2)
        # print(repr(t))
        # print(repr(t2))

    def test_time_pkt_2(self):
        t = ch10time.TimeDataFormat2()
        t.ptptime = PTPTime(1558366884, 999999999)
        # t.nanoseconds = 999999999
        # t.datetime = datetime.datetime.fromtimestamp(1558366884)

        t2 = ch10time.TimeDataFormat2()
        t2.unpack(t.pack())
        self.assertTrue(t2 == t)
        # print(repr(t))
        # self.assertEqual("TimeFormat2 ChannelSpecificWord=0X11 Time=16:41:24 05/20/19 20-May 2019 NanoSeconds=999999999", repr(t))
        # print(repr(t2))

    def test_time_pkt_1_decom(self):
        t = ch10time.TimeDataFormat1()
        # 0000   11 00 00 00 99 09 13 14 06 02
        # Tue, 25 Jul 2023 14:13:09 GMT
        t.unpack(struct.pack(">HHHHH", 0x1100, 0x0, 0x9909, 0x1314, 0x0602))
        _t = repr(t)
        ref_t = PTPTime(17763189, 990_000_000)
        self.assertEqual(t.ptptime, ref_t)
        # self.assertEqual(t.nanoseconds, 990_000_000)
        self.assertEqual(
            "TimeFormat1 ChannelSpecificWord=0X11 Time=PTP: 14:13:09 25-Jul 1970 nanosec=990000000", repr(t)
        )

    def test_time_to_pcap(self):
        pcapw = pcap.Pcap(TMP_DIR + "/test_ch10_time.pcap", mode="w")

        rec = pcap.PcapRecord()
        cu = ch10udp.Chapter10UDP()
        cu.type = ch10udp.Chapter10UDP.TYPE_FULL
        cu.sequence = 1
        c = get_ch10()
        types = [DATA_TYPE_TIMEFMT_1, DATA_TYPE_TIMEFMT_2]
        t1 = ch10time.TimeDataFormat1()
        t2 = ch10time.TimeDataFormat2()
        for i, t in enumerate([t1, t2]):
            t.ptptime = PTPTime(1558366884, 980_000_000)
            c.payload = t.pack()
            c.datatype = types[i]
            full_payload = getEthernetPacket(cu.pack())
            rec.payload = full_payload
            self.assertIsNone(pcapw.write(rec))
        pcapw.close()

    @unittest.skip("")
    def test_ch10_fmt_pcap(self):
        p = pcap.Pcap(os.path.join(THIS_DIR, "ch10_pkt.pcap"))
        mypcaprecord = p.next()
        c = ch10udp.Chapter10UDP()
        c.unpack(mypcaprecord.payload[0x2A:])
        t = ch10time.TimeDataFormat1()
        t.unpack(c.chapter10.payload)
        # print(t.datetime.strftime("%c"))
        mypcaprecord = p.next()
        c = ch10udp.Chapter10UDP()
        c.unpack(mypcaprecord.payload[0x2A:])
        t = ch10time.TimeDataFormat2()
        t.unpack(c.chapter10.payload)
        # print(t.datetime.strftime("%c"))
        self.assertEqual(t.datetime, datetime.datetime.fromtimestamp(1561535306))


class GenPCAP(unittest.TestCase):
    def test_ch10_fmt1(self):
        # self.assertEqual(TMP_DIR, "")
        pcapw = pcap.Pcap(TMP_DIR + "/test_ch10_sample.pcap", mode="w")

        rec = pcap.PcapRecord()
        for idx in range(3):
            c = get_ch10(16)
            c.ptptime = PTPTime(333344, 42322)
            c.packetflag = (
                ch10.Chapter11.PKT_FLAG_SECONDARY
                + ch10.Chapter11.PKT_FLAG_SEC_HDR_TIME
                + ch10.Chapter11.PKT_FLAG_1588_TIME
            )
            cu = ch10udp.Chapter10UDP()
            cu.type = ch10udp.Chapter10UDP.TYPE_FULL
            cu.sequence = 1
            # format 2
            cu.segmentoffset = 0x0
            cu.channelID = c.channelID
            # Format 3
            cu.sourceid_len = 0x1
            cu.sourceid = 0x2
            cu.sequence = 0x3
            cu.offset_pkt_start = 0x0

            cu.format = 1

            # Write it out

            full_payload = getEthernetPacket(cu.pack() + c.pack())
            rec.payload = full_payload
            self.assertIsNone(pcapw.write(rec))
        pcapw.close()


class Ch10Mil(unittest.TestCase):
    def test_mil(self):
        m = ch10mil.MILSTD1553DataPacket(TS_IEEE1558)
        msg = ch10mil.MILSTD1553Message()
        msg.message = struct.pack(">II", 1, 2)
        msg.ipts = PTPTime(100, 200)
        m.append(msg)
        m2 = ch10mil.MILSTD1553DataPacket(TS_IEEE1558)
        m2.unpack(m.pack())
        self.assertEqual(m2, m)


class PCMData(unittest.TestCase):
    def test_pcm(self):
        payload_len = 100
        sync_word = 0xFE6B_2842
        pcapw = pcap.Pcap(TMP_DIR + "/test_ch10_pcm.pcap", mode="w")
        rec = pcap.PcapRecord()
        u = ch10udp.Chapter10UDP()
        c = ch10.Chapter11()
        u.type = ch10udp.Chapter10UDP.TYPE_FULL
        u.sequence = 0
        c = get_ch10()
        c.datatype = 0x9
        pcmdf = ch10pcm.PCMDataPacket()
        pcmdf.channel_specific_word = 0x7F080000
        for mfcount in range(1, 4):
            mf = ch10pcm.PCMMinorFrame()
            mf.intra_packet_data_header = mfcount
            mf.ipts.count = 2 * mfcount
            mf.minor_frame_data = struct.pack(">I", sync_word) + os.urandom(payload_len - 4)
            pcmdf.minor_frames.append(mf)
        packed_data = pcmdf.pack()
        c.payload = packed_data
        u.payload = c.pack()
        rec.set_current_time()
        rec.payload = getEthernetPacket(u.pack())
        pcapw.write(rec)
        pcapw.close()

        pcmdf2 = ch10pcm.PCMDataPacket(syncword=sync_word)
        # pcmdf2.minor_frame_size_bytes = payload_len
        pcmdf2.unpack(packed_data)
        self.assertEqual(pcmdf, pcmdf2)
        pcmdf3 = ch10pcm.PCMDataPacket(syncword=None)
        self.assertTrue(pcmdf3.unpack(packed_data))
        # print(pcmdf3)

    def test_pcm_throughput(self):
        pcmdf = ch10pcm.PCMDataPacket()
        pcmdf.channel_specific_word = 0x100000  #
        mf = ch10pcm.PCMMinorFrame(throughput=True)
        mf.minor_frame_data = struct.pack("<HHI", 0xFE6B, 0x2840, 0)
        pcmdf.minor_frames.append(mf)
        packed = pcmdf.pack()
        pcmdf2 = ch10pcm.PCMDataPacket()
        pcmdf2.unpack(packed)
        self.assertEqual(pcmdf, pcmdf2)
        self.assertEqual(
            repr(pcmdf2),
            "PCM Data Packet Format 1. Channel Specific Word =0X100000\nMinor Frame Throughput mode Time=None Payload_len=8\n",
        )
        print(base64.b64encode(packed))

    def test_pcm_endianness(self):
        mf = ch10pcm.PCMMinorFrame(throughput=True)
        mf.minor_frame_data = struct.pack("<I", 0xFE6B2840) + struct.pack(">I", 0xFE6B2840)
        f = open("mf.hex", mode="wb")
        f.write(mf.pack())
        f.close()


class MnACQData(unittest.TestCase):
    def test_pcap(self):
        p = pcap.Pcap(os.path.join(THIS_DIR, "mnacq2.pcap"))
        ch10_payload = b""
        cf = open("{}/mnacq.csv".format(THIS_DIR), mode="w", newline="")
        csvf = csv.writer(cf)

        for pkt_count, rec in enumerate(p):
            wrapper = ch10udp.Chapter10UDP()
            pkt = ch10.Chapter11()
            wrapper.unpack(rec.payload[0x2A:])

            if wrapper.type == ch10udp.Chapter10UDP.TYPE_SEG and wrapper.segmentoffset == 0 and ch10_payload != b"":
                # segmented packet with existing payload
                self.assertTrue(pkt.unpack(ch10_payload))
                # print(repr(pkt))
                ch10_payload = b""
                self.assertEqual(pkt.datatype, DATA_TYPE_PCM_DATA_FMT1)
                pcm = ch10pcm.PCMDataPacket()
                pcm.minor_frame_size_bytes = 64
                try:
                    s = pcm.unpack(pkt.payload, extract_sync_sfid=False)
                except Exception as e:
                    print("At packet count {}. failed to unpack pcmd data. Err={}".format(pkt_count + 1, e))
                    self.assertTrue(False)
                else:
                    self.assertTrue(s)
                first_minor_frame = pcm.minor_frames[0]
                if pkt_count == 62:
                    self.assertEqual(68676670524492, first_minor_frame.ipts.count)
                for mf in pcm.minor_frames:
                    num_words = int(len(mf.minor_frame_data) / 2)
                    dw = struct.unpack(f"<{num_words}H", mf.minor_frame_data)
                    csvf.writerow(list(dw))

            if wrapper.type == ch10udp.Chapter10UDP.TYPE_SEG:
                ch10_payload += wrapper.payload
            else:
                pkt.unpack(wrapper.payload)
                # print(repr(pkt))

        cf.close()
        p.close()


class CH10SampleFile(unittest.TestCase):
    @unittest.skip("sample file too big at the moment")
    def test_dotch10(self):
        fileparser = ch10.FileParser(THIS_DIR + "/ch10.ch10")
        total_len = 0
        with fileparser as chf:
            for pkt in chf:
                # print(repr(pkt))
                total_len += pkt.packetlen
                # self.assertTrue(False)
        # print(f"{total_len:,d}")

    @unittest.skip("Not a test")
    def test_read_ch10(self):
        fileparser = ch10.FileParser("C://ACRA//WORK//AcraNetwork_git//examples//ch10//out.ch10")
        total_len = 0
        with fileparser as chf:
            for pkt in chf:
                pkt_repr = repr(pkt)
                total_len += pkt.packetlen
                # self.assertTrue(False)
        # print(f"{total_len:,d}")

    @unittest.skip("Not a test")
    def test_extract_tmats(self):
        fileparser = ch10.FileParser("C://ACRA//WORK//AcraNetwork_git//examples//ch10//Kulite Recording.ch10")
        tmats_f = open("C://ACRA//WORK//AcraNetwork_git//examples//ch10//Kulite.tmats", mode="wb")
        with fileparser as chf:
            for idx, pkt in enumerate(chf):
                if idx == 0:
                    tmats_f.write(pkt.payload[4:])
                    break


class PTPRTCTime(unittest.TestCase):
    def test_ptptime(self):
        t0 = PTPTime(0, 0)
        t0b = PTPTime(0, 0)
        t1 = PTPTime(2, 1)
        t2 = PTPTime(1, 999_999_999)
        t2b = PTPTime(1, 999_999_999)
        dlt = PTPTime(0, 22)
        t3 = PTPTime(1, 0)

        self.assertTrue(t0 == t0b)
        self.assertTrue(t1 > t0)
        self.assertTrue(t2 < t1)
        self.assertTrue(t0 >= t0)
        self.assertTrue(t0 <= t0)
        self.assertEqual(t1 + t2, PTPTime(4, 0))
        self.assertEqual(t1 - t2, PTPTime(0, 2))
        self.assertEqual(t2 + t2, PTPTime(3, 999_999_998))
        self.assertEqual(t2 - t2b, PTPTime(0, 0))
        self.assertFalse(dlt >= t3)

    def test_ptptime_conversion(self):
        F8BITS = pow(2, 48) - 1
        t0 = PTPTime(0, 0)
        self.assertEqual(t0.to_pinksheet_rtc(), 0)
        self.assertEqual(t0.to_rtc(), 0)
        t1 = PTPTime(4096, 1)

        time_vector = struct.pack(">II", 4096, 1)
        (sec, nsec) = struct.unpack(">II", time_vector)
        trunc = int((sec * 1e9 + nsec) // 100) & F8BITS
        self.assertEqual(t1.to_pinksheet_rtc(), trunc)

    def test_accuracy_conversion(self):
        t0 = PTPTime(1704299074, 472711723)
        rtc_time = int(1.54492142087757e14)
        self.assertEqual(rtc_time, t0.to_pinksheet_rtc())


class Video(unittest.TestCase):
    def test_video(self):
        vid = ch10video.VideoFormat2()
        vid.channel_specific_word = 0x0
        mpgsts = ampeg.MPEGTS()
        for _i in range(1):
            mpegp = ampeg.MPEGPacket()
            mpegp.sync = 0x47
            mpegp.adaption_ctrl = ampeg.ADAPTION_PAYLOAD_ONLY
            mpegp.pid = _i
            mpegp.payload = struct.pack(f"{92}H", *range(92))
            mpgsts.append(mpegp)
        vid.mpegts = mpgsts
        buf = vid.pack()
        # print(repr(buf))
        vid2 = ch10video.VideoFormat2()
        vid2.unpack(buf)
        self.assertEqual(vid, vid2)


class CustomException(Exception):
    def __init__(self, message=""):
        print(f"Exception : {message}")


class ExceptionOverride(unittest.TestCase):

    @unittest.skip("Not supported")
    def test_cillian(self):
        p = pcap.Pcap(os.path.join(THIS_DIR, "CH10_CW_Extract_20250313.pcap"))
        mypcaprecord = p[0]
        p.close()
        _ch10pkt = ch10udp.Chapter10UDP()
        _ch10pkt.unpack(mypcaprecord.payload[0x2A:])


if __name__ == "__main__":
    unittest.main()



import struct
import logging

logger = logging.getLogger(__name__)


class Chapter10UDP(object):
    """
    Class to encapsulate Chapter10 payload in UDP packets

    Capture a UDP packet and unpack the payload as an Chapter 10 UDP packet. The payload can then be passed to Chapter10 class

    There are two types of packets, segmented and full.

    >>> import AcraNetwork.Pcap as pcap
    >>> p = pcap.Pcap("test/ch10_uart2.pcap")
    >>> data = p[3].payload[0x2A:]
    >>> n = Chapter10UDP()
    >>> n.unpack(data)
    True
    >>> print(f"{n.format}. Seq={n.sequence}")
    1. Seq=3


    """

    CH10_UDP_HEADER_FORMAT1 = "<BBH"
    CH10_UDP_HEADER_FORMAT2 = ">HBB"
    CH10_UDP_SEG_HEADER_FORMAT1 = "<HBBI"

    CH10_HDR_LEN = {1: 4, 2: 12, 3: 8}  # fmt1, 2, 3

    CH10_UDP_HEADER_LENGTH = struct.calcsize(CH10_UDP_HEADER_FORMAT1)
    CH10_UDP_SEG_HEADER_LENGTH = struct.calcsize(CH10_UDP_SEG_HEADER_FORMAT1)

    TYPE_FULL = 0  #: Full Chapter 10 packets type field constant. Assign to :attr:`Chapter10UDP.type`
    TYPE_SEG = 1  #: Segmented Chapter 10 packets type field constant. Assign to :attr:`Chapter10UDP.type`

    def __init__(self):
        """Creator method for a UDP class"""
        self.version = 1  #: Version
        self.type = Chapter10UDP.TYPE_FULL  #: Type of message , Full or Segmented
        self.channelID = 0  #: Segmented Packets Only. Channel ID of the data in the RCC 106 Chapter 10 packet
        self.channelsequence = 0  #: Segmented Packets Only,
        self.sequence = 0  #: UDP Sequence number
        self.segmentoffset = 0  #: Segmented Packets Only. The 32-bit Segmented Packets Only, Position of the data in the RCC 106 Chapter 10 packet.
        self.packetsize = None  #: Format 2 Packet size
        self.sourceid_len = 0  #: Format 3 Source ID length
        self.sourceid = 0  #: Format 3 Source ID
        self.offset_pkt_start = None  #: Format 3 Offset to packet start in bytes
        self.payload = b""

    def unpack(self, buffer):
        """
        Unpack a string buffer into an Chapter10UDP object

        :param buffer: A string buffer representing an Chapter10UDP packet
        :type buffer: bytes
        :rtype: None
        """
        (_ver_type, seg_lwr, seg_upr) = struct.unpack_from(Chapter10UDP.CH10_UDP_HEADER_FORMAT1, buffer)
        if _ver_type & 0xF == 1:
            self.version = 1
            self.type = _ver_type >> 4
            self.sequence = seg_lwr + (seg_upr << 8)
        elif _ver_type & 0xF == 3:
            self.version = 3
            self.type = _ver_type >> 4
        else:
            self.version = 2
            (seg_upr, seg_lwr, _ver_type) = struct.unpack_from(Chapter10UDP.CH10_UDP_HEADER_FORMAT2, buffer)
            (
                _size_upp,
                _size_lower,
            ) = struct.unpack_from(">BH", buffer, 5)
            self.type = _ver_type >> 4
            self.sequence = seg_lwr + (seg_upr << 8)

        if self.version == 3:
            self.sourceid_len = _ver_type >> 4
            self.offset_pkt_start = seg_upr

        if self.type == Chapter10UDP.TYPE_SEG and self.format == 1:
            raise Exception("No Supported")
            (self.channelID, self.channelsequence, _res, self.segmentoffset) = struct.unpack_from(
                Chapter10UDP.CH10_UDP_SEG_HEADER_FORMAT1, buffer, Chapter10UDP.CH10_UDP_HEADER_LENGTH
            )
            self.payload = buffer[(Chapter10UDP.CH10_UDP_HEADER_LENGTH + Chapter10UDP.CH10_UDP_SEG_HEADER_LENGTH) :]
        elif self.format == 1:
            self.payload = buffer[Chapter10UDP.CH10_UDP_HEADER_LENGTH :]
        elif self.format == 2:
            (_segoffset_upper, _size_upp, _size_lower, _segoff_lower, self.channelID) = struct.unpack_from(
                ">BBHHH", buffer, 4
            )
            self.packetsize = _size_lower + (_size_upp << 16)
            self.segmentoffset = _segoff_lower + (_segoffset_upper << 16)
            self.payload = buffer[Chapter10UDP.CH10_UDP_HEADER_LENGTH + 8 :]
        elif self.format == 3:
            (_srcid_datat,) = struct.unpack_from("<I", buffer, 4)
            if self.sourceid_len == 0:
                self.sourceid = 0x0
                self.sequence = _srcid_datat
            elif self.sourceid_len == 1:
                self.sourceid = _srcid_datat >> (32 - 4)
                self.sequence = _srcid_datat & 0x0FFFFFFF
            elif self.sourceid_len == 2:
                self.sourceid = _srcid_datat >> (32 - 8)
                self.sequence = _srcid_datat & 0x00FFFFFF
            elif self.sourceid_len == 3:
                self.sourceid = _srcid_datat >> (32 - 12)
                self.sequence = _srcid_datat & 0x000FFFFF
            elif self.sourceid_len == 4:
                self.sourceid = _srcid_datat >> (32 - 16)
                self.sequence = _srcid_datat & 0x0000FFFF
            else:
                raise Exception("Source id length {} is not valid".format(self.sourceid_len))

            self.payload = buffer[Chapter10UDP.CH10_UDP_HEADER_LENGTH + 4 :]

        else:
            self.payload = buffer[Chapter10UDP.CH10_UDP_HEADER_LENGTH :]

        return True

    @property
    def format(self):
        return self.version

    @format.setter
    def format(self, val):
        self.version = val

    def pack(self):
        """
        Pack the Chapter10UDP object into a binary buffer

        :rtype: bytes
        """

        if self.format == 3:
            _ver_type = (self.sourceid_len << 4) + self.version
            seg_up = self.offset_pkt_start
            seg_lr = 0
        else:
            _ver_type = (self.type << 4) + self.version
            seg_up = self.sequence >> 8
            seg_lr = self.sequence & 0xFF

        if self.format == 2:
            _payload = struct.pack(Chapter10UDP.CH10_UDP_HEADER_FORMAT2, seg_up, seg_lr, _ver_type)
        else:
            _payload = struct.pack(Chapter10UDP.CH10_UDP_HEADER_FORMAT1, _ver_type, seg_lr, seg_up)

        if self.type == Chapter10UDP.TYPE_SEG and self.format == 1:
            _payload += struct.pack(
                Chapter10UDP.CH10_UDP_SEG_HEADER_FORMAT1, self.channelID, self.channelsequence, 0, self.segmentoffset
            )

        elif self.format == 2:
            self.packetsize = len(self.payload) // 4
            _payload += struct.pack(
                ">BBHHH",
                self.segmentoffset >> 16,
                self.packetsize >> 16,
                self.packetsize & 0xFFFF,
                self.segmentoffset & 0xFFFF,
                self.channelID,
            )

        elif self.format == 3:
            if self.sourceid_len == 0:
                _field = self.sequence
            elif self.sourceid_len == 1:
                _field = (self.sequence & 0x0FFFFFFF) + (self.sourceid << (32 - 4))
            elif self.sourceid_len == 2:
                _field = (self.sequence & 0x00FFFFFF) + (self.sourceid << (32 - 8))
            elif self.sourceid_len == 3:
                _field = (self.sequence & 0x000FFFFF) + (self.sourceid << (32 - 12))
            elif self.sourceid_len == 4:
                _field = (self.sequence & 0x0000FFFF) + (self.sourceid << (32 - 16))
            else:
                _field = 0
                raise Exception("Invalid source id")

            _payload += struct.pack("<I", _field)

        return _payload + self.payload

    def __repr__(self):
        if self.type == Chapter10UDP.TYPE_FULL:
            return "CH10 UDP Full Packet: Format={} Sequence={}".format(self.format, self.sequence)
        else:
            return "CH10 UDP Sequence: Format={} Sequence={} ChID={} ChSeqNum={} SegOffset={}".format(
                self.format, self.sequence, self.channelID, self.channelsequence, self.segmentoffset
            )

    def __eq__(self, other):
        if not isinstance(other, Chapter10UDP):
            return False

        if self.type == Chapter10UDP.TYPE_SEG and self.format == 1:
            _match_att = ("version", "type", "sequence", "channelID", "channelsequence", "segmentoffset", "payload")
        elif self.format == 2:
            _match_att = (
                "format",
                "type",
                "sequence",
                "channelID",
                "channelsequence",
                "segmentoffset",
                "packetsize",
                "payload",
            )
        elif self.format == 3:
            _match_att = ("format", "sourceid_len", "sourceid", "sequence", "offset_pkt_start", "payload")
        else:
            _match_att = ("version", "type", "sequence", "payload")

        for attr in _match_att:
            if getattr(self, attr) != getattr(other, attr):
                logger.debug(
                    f"Attributes {attr} does not match. self={getattr(self, attr)} other={getattr(other, attr)}"
                )
                return False

        return True



import struct
import typing
import logging
from AcraNetwork.MPEGTS import MPEGPacket
from functools import reduce
import datetime

logger = logging.getLogger(__name__)


PES_EXTENSION_W1 = 0x81
PES_EXTENSION_W2 = 0x80


def pts_to_ts(v: int) -> float:
    """Presentation Timestamp to seconds floating"""
    pts = 0
    pts |= (v >> 3) & (0x0007 << 30)  # // top 3 bits, shifted left by 3, other bits zeroed out
    pts |= (v >> 2) & (0x7FFF << 15)  # // middle 15 bits
    pts |= (v >> 1) & (0x7FFF << 0)  # // bottom 15 bits
    return pts / 90e3


def ts_to_pts(ts: float) -> int:
    """Floating point seconds to PTS"""
    pts = int(ts * 90e3)
    v = 0x2100010001  # I have no idea where the 2 comes from at the msb
    v |= (pts & 0x7FFF) << 1  # // bottom 15 bits
    v |= ((pts >> 15) & 0x7FFF) << 17  # // middle 15 bits
    v |= ((pts >> 30) & 0x7) << 31  # // top 3 bits

    return v


def ts_to_buf(ts: float) -> bytes:
    v = ts_to_pts(ts)
    return struct.pack(">BI", (v >> 32), v & 0xFFFF_FFFF)


def buf_to_ts(buffer: bytes) -> float:
    """Convert a buffer (ie PES header) to a timestamp"""
    (msb, lsb) = struct.unpack(">BI", buffer)
    pts_ts = lsb + (msb << 32)
    return pts_to_ts(pts_ts)


class PES(MPEGPacket):

    def __init__(self) -> None:
        super().__init__()
        self.streamid: int = 0
        self.pesdata: bytes = bytes()
        self.extension_w1: typing.Optional[int] = None
        self.extension_w2: typing.Optional[int] = None
        self.header_data: typing.Optional[bytes] = None

    def unpack(self, buffer: bytes):
        super().unpack(buffer)
        (_prefix1, _prefix2, self.streamid, _peslength) = struct.unpack_from(">BHBH", self.payload)
        prefix = (_prefix1 << 16) + _prefix2
        if prefix != 1:
            raise Exception(f"PES Prefix {prefix:#0X} should be 0x1")
        # Peek to see if the data is an extension
        (optional_hdr, _miscbits, _pes_hdr_len) = struct.unpack_from(">BBB", self.payload, 6)
        marker = optional_hdr >> 4
        expected_len_if_extension_present = _peslength + 6
        if marker == 0x8 and len(self.payload) == expected_len_if_extension_present:
            # PES extenion
            (self.extension_w1, self.extension_w2, _hdrlen) = struct.unpack_from(">BBB", self.payload, 6)
            self.header_data = self.payload[9 : (9 + _hdrlen)]
            self.pesdata = self.payload[(9 + _hdrlen) :]
        else:
            logger.debug("No optional PES header")
            self.pesdata = self.payload[6:]

    def pack(self) -> bytes:

        if self.extension_w1 is not None and self.extension_w2 is not None and self.header_data is not None:
            _len = 3 + len(self.pesdata) + len(self.header_data)
            _ext_present = True
        else:
            _ext_present = False
            _len = len(self.header_data)
        self.payload = struct.pack(">BHBH", 0, 1, self.streamid, _len)
        if _ext_present:
            self.payload += struct.pack(">BBB", self.extension_w1, self.extension_w2, len(self.header_data))
            self.payload += self.header_data
        self.payload += self.pesdata

        return super().pack()

    def __repr__(self) -> str:
        r = super().__repr__()
        if self.extension_w1 is None:
            _ext = ""
        else:
            _ext = "Extension"
        return r + "\n" f"   PES {_ext}: Stream ID={self.streamid:#0X}  PESData Len={len(self.pesdata)}"

    def __eq__(self, __value: object) -> bool:
        if self.streamid != __value.streamid or self.pesdata != __value.pesdata:
            return False
        if (
            self.extension_w1 != __value.extension_w1
            or self.extension_w2 != __value.extension_w2
            or self.header_data != __value.header_data
        ):
            return False
        return super().__eq__(__value)


def checksum_stanag(buff: bytes) -> int:
    """Calculate the STANAG checksum as described in 6.8
    https://upload.wikimedia.org/wikipedia/commons/1/19/MISB_Standard_0601.pdf

    """
    bcc = 0

    # Sum each 16-bit chunk within the buffer into a checksum
    for i in range(len(buff)):
        bcc += buff[i] << (8 * ((i + 1) % 2))

    return bcc % 65536


STANAG4609_UNIVERSAL_KEY = struct.pack(">QQ", 0x060E2B34020B0101, 0x0E01030101000000)
STANAG4609_UNKNOWN_OFFSET = 5  # There's some data at the start that I don't understand
STANAG4609_DATA_TAG = 0x2
STANAG4609_TIME_TAG = 0x1
STANAG4609_PID = 0x104
STANAG4609_LEN = 0xE
STANAG4609_DTAG_LEN = 0x8
STANAG4609_TTAG_LEN = 0x2


class STANAG4609(PES):

    def __init__(self) -> None:
        super().__init__()
        self.stanag_counter: int = 0
        self._unknown: int = 0xDF
        self._unknown2: int = 0x1F

        self.time_us: int = 0  # Store the time in us

    # Set the time as a date time option
    @property
    def time(self):
        return datetime.datetime.fromtimestamp(self.time_us / 1e6)

    @time.setter
    def time(self, value: datetime.datetime):
        self.time_us = int(value.timestamp() * 1e6)

    def unpack(self, buffer: bytes):
        """Convert the buffer into a STANAG4609 object"""
        super().unpack(buffer)
        if self.pid != STANAG4609_PID:
            raise Exception(f"PID of STANAG should be 0x104 not {self.pid:#0X}")
        # I can't find documentation for these 3 fields. One looks like a counter
        (self.stanag_counter, self._unknown, self._unknown2) = struct.unpack_from(">HBH", self.pesdata, 0)
        _offset = len(STANAG4609_UNIVERSAL_KEY) + STANAG4609_UNKNOWN_OFFSET
        # Check the universal key
        if self.pesdata[STANAG4609_UNKNOWN_OFFSET:_offset] != STANAG4609_UNIVERSAL_KEY:
            raise Exception("STANAG4609 requires a specific UNIVERSAL KEY")
        # Most of these are static
        (_len, _datatag, _tag_len) = struct.unpack_from(">BBB", self.pesdata, _offset)
        if _datatag != STANAG4609_DATA_TAG:
            raise Exception("Data tag of STANAG should be 0x2")
        if _tag_len != 8:
            raise Exception(f"Tag length should be 8 not {_tag_len}")
        _offset += 3
        # Pull out the time and store as a datetime
        (self.time_us, _timetag, _time_len, actchecksum) = struct.unpack_from(">QBBH", self.pesdata, _offset)
        checksum = checksum_stanag(self.pesdata[STANAG4609_UNKNOWN_OFFSET:-2])
        if checksum != actchecksum:
            raise Exception(f"Extracted checksum={actchecksum:#0X} Calculated={checksum:#0X}")

    def pack(self) -> bytes:
        """Pack the STANAG4609 object into bytes

        Returns:
            bytes: _description_
        """
        self.pid = STANAG4609_PID
        self.pesdata = struct.pack(">HBH", self.stanag_counter, self._unknown, self._unknown2)  # The unknown header
        self.pesdata += STANAG4609_UNIVERSAL_KEY + struct.pack(
            ">BBB", STANAG4609_LEN, STANAG4609_DATA_TAG, STANAG4609_DTAG_LEN
        )
        self.pesdata += struct.pack(">Q", self.time_us) + struct.pack(">BB", STANAG4609_TIME_TAG, STANAG4609_TTAG_LEN)
        checksum = checksum_stanag(self.pesdata[STANAG4609_UNKNOWN_OFFSET:])
        self.pesdata += struct.pack(">H", checksum)
        # _lp = len(self.pesdata)
        return super().pack()

    def __repr__(self) -> str:
        r = super().__repr__()
        time_fmt = datetime.datetime.fromtimestamp(self.time_us / 1e6)
        r += "\n" + f"   Time={repr(time_fmt)}"
        return r

    def __eq__(self, __value: object) -> bool:
        if self.time_us != __value.time_us or self.stanag_counter != __value.stanag_counter:
            return False
        return super().__eq__(__value)



__author__ = "diarmuid"
import sys

sys.path.append("..")
import os

import unittest
import AcraNetwork.Pcap as pcap
import AcraNetwork.SimpleEthernet as SimpleEthernet
import struct


THIS_DIR = os.path.dirname(os.path.abspath(__file__))


def getEthernetPacket(data=0xA):
    e = SimpleEthernet.Ethernet()
    e.srcmac = 0x001122334455
    e.dstmac = 0x998877665544
    e.type = SimpleEthernet.Ethernet.TYPE_IP
    e.payload = struct.pack("H", data)
    return e.pack()


class PcapBasicTest(unittest.TestCase):
    def test_missingfilename(self):
        self.assertRaises(TypeError, lambda: pcap.Pcap())

    def test_missingreadfile(self):
        self.assertRaises(IOError, lambda: pcap.Pcap(os.path.join(THIS_DIR, "nofile.pcap")))

    def test_defaultMagicNumber(self):
        p = pcap.Pcap("_tmp.pcap", mode="w")
        self.assertEqual(p.magic, 0xA1B2C3D4)
        p.close()

    def test_defaultVersionMaj(self):
        p = pcap.Pcap("_tmp.pcap", mode="w")
        self.assertEqual(p.versionmaj, 2)
        p.close()

    def test_defaultVersionMin(self):
        p = pcap.Pcap("_tmp.pcap", mode="w")
        self.assertEqual(p.versionmin, 4)
        p.close()

    def test_readTestFile(self):
        p = pcap.Pcap(os.path.join(THIS_DIR, "test_input.pcap"))
        self.assertEqual(p.magic, 0xA1B2C3D4)
        self.assertEqual(p.network, 1)
        self.assertEqual(p.sigfigs, 0)
        self.assertEqual(p.snaplen, 262144)
        self.assertEqual(p.versionmaj, 2)
        self.assertEqual(p.versionmin, 4)
        self.assertEqual(p.zone, 0)
        self.assertEqual(p.filesize, 704)
        for idx, rec in enumerate(p):
            if id == 3:
                self.assertEqual(len(rec), 66)
        p.close()

    def test_readARecord(self):
        p = pcap.Pcap(os.path.join(THIS_DIR, "test_input.pcap"))
        mypcaprecord = p[0]
        p.close()
        self.assertEqual(mypcaprecord.sec, 1419678111)
        self.assertEqual(mypcaprecord.usec, 811463)
        self.assertEqual(mypcaprecord.orig_len, 70)
        self.assertEqual(mypcaprecord.incl_len, 70)
        self.assertEqual(repr(mypcaprecord), "LEN:70 SEC:1419678111 USEC:811463")
        p.close()

    def test_writeARecord(self):
        p = pcap.Pcap("_tmp.pcap", mode="w")
        r = pcap.PcapRecord()
        r.setCurrentTime()
        r.packet = getEthernetPacket(0xA)
        p.write(r)
        p.close()
        p = pcap.Pcap("_tmp.pcap")
        self.assertEqual(p.magic, 0xA1B2C3D4)
        self.assertEqual(p.network, 1)
        self.assertEqual(p.sigfigs, 0)
        self.assertEqual(p.snaplen, 65535)
        self.assertEqual(p.versionmaj, 2)
        self.assertEqual(p.versionmin, 4)
        self.assertEqual(p.zone, 0)
        self.assertEqual(p.filesize, 56)
        p.close()
        os.remove("_tmp.pcap")

    def test_appendARecord(self):
        p = pcap.Pcap("_tmp2.pcap", mode="w")
        r = pcap.PcapRecord()
        r.setCurrentTime()
        r.packet = getEthernetPacket(0xA)
        p.write(r)
        p.close()
        # Now try to append a record
        p = pcap.Pcap("_tmp2.pcap", mode="a")
        r.packet = getEthernetPacket(0xB)
        p.write(r)
        p.close()
        # Read back to verify
        p = pcap.Pcap("_tmp2.pcap")
        self.assertEqual(p.filesize, 88)
        for idx, rec in enumerate(p):
            e = SimpleEthernet.Ethernet()
            e.unpack(rec.packet)
            if idx == 0:
                self.assertEqual(e.payload, struct.pack("H", 0xA))
            else:
                self.assertEqual(e.payload, struct.pack("H", 0xB))
        p.close()
        os.remove("_tmp2.pcap")

    def test_context_manager(self):
        """
        tests the use of __enter__() and __exit() to allow "with" to ensure
        the file is closed. Also verifies the behaviour of the rec_no value.        
        """
        test_filename = "_tmp3.pcap"

        # define a payload for packet i
        def pkt_payload(i): return i+1234

        with pcap.Pcap(test_filename, mode="w") as p:
            for i in range(10):
                r = pcap.PcapRecord(now=True)
                r.packet = getEthernetPacket(pkt_payload(i))
                p.write(r)
                self.assertEqual(p.rec_no, i+1)
        self.assertTrue(p.fopen.closed)
        
        # Read the file back. This also shows how to handle an exception when 
        # opening the file... the first filename will be invalid, the second 
        # will work.
        for filename in ("this_file_does_not_exist", test_filename):
            exception_msg = None
            if filename == test_filename:
                exp_exception_msg = None
            else:
                exp_exception_msg = f"Failed to open {filename}. err=[Errno 2] No such file or directory: '{filename}'"
            try:
                p = pcap.Pcap(filename, mode="r")
            except IOError as e:
                exception_msg = str(e)
            else:
                with p:
                    for i, rec in enumerate(p):
                        e = SimpleEthernet.Ethernet()
                        e.unpack(rec.packet)

                        self.assertEqual(e.payload, 
                                         struct.pack("H", pkt_payload(i))
                                        )
            self.assertEqual(exception_msg, exp_exception_msg)
            if filename == test_filename:
                # file should have been closed properly by __exit__()
                self.assertTrue(p.fopen.closed)
            else:
                # exception should have happened so object has not even been
                # created
                self.assertIsNot(exception_msg, None)
        os.remove(test_filename)
        

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

# Requires cryptography and zstandard
# pip install zstandard
# pip install cryptography
# -*- coding: utf-8 -*-


import AcraNetwork.IRIG106.Chapter11 as ch11
import AcraNetwork.IRIG106.Chapter11.TimeDataFormat as chtime
from AcraNetwork.IRIG106.Chapter10 import FileParser
import AcraNetwork.IRIG106.Chapter10.Chapter10UDP as ch10udp
from AcraNetwork.IRIG106.Chapter11 import PTPTime
import AcraNetwork.Pcap as pcap
import AcraNetwork.SimpleEthernet as eth
import argparse
import glob
import os.path
from dataclasses import dataclass, field
import typing
import sys
import logging
import time
from functools import reduce
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization
from pathlib import Path
import tempfile
from os import remove
from shutil import rmtree
import struct
import zstandard


RED = "\033[31m"
RESET = "\033[0m"
TIME_CHID = 0x1
DATA_TYPE_WRAPPED_ETHERNET = 0x68

logging.basicConfig(level=logging.INFO)

# This will make the validation 10x slower. It validates all the data in the inetx packet
CHECK_INETX_PAYLOAD = False
INETX_PAYLOAD_LEN_WORDS = 722


@dataclass
class Channels:
    channel: int
    sequence: int
    pkt_count: int
    dropcnt: int
    datavol: int

    def __repr__(self):
        if self.dropcnt > 0:
            hi = RED
        else:
            hi = ""
        return f"ChannelID={self.channel:#08X} Count={self.pkt_count:12d} {hi}Drop={self.dropcnt:10d}{RESET}, Vol={self.datavol / 1e6:10.1f}MB"


def create_parser():
    # ----------------------------------
    # Setup the command line parser
    # ----------------------------------
    parser = argparse.ArgumentParser(description="Validate a ch10 recording")
    parser.add_argument("--folder", type=str, required=True, default=None, help="folder to ch10 files")
    parser.add_argument(
        "--unwrap", type=str, required=False, default=None, help="folder to write unwrapped ethernet packets"
    )
    parser.add_argument("--key", type=str, required=False, default=False, help="RSA private key")
    parser.add_argument(
        "--decompress", action="store_true", required=False, default=False, help="decompress the file before validation"
    )

    return parser


def decrypt_aes_key(ciphertext: bytes, pem_file: str) -> bytes:
    # Load the private key from a PEM file
    with open(pem_file, "rb") as priv_file:
        private_pem = priv_file.read()

    private_key = serialization.load_pem_private_key(
        private_pem, password=None  # Provide a password if the private key is encrypted
    )
    # Decrypt the data using the private key
    decrypted_data = private_key.decrypt(
        ciphertext, padding.OAEP(mgf=padding.MGF1(algorithm=hashes.SHA1()), algorithm=hashes.SHA1(), label=None)
    )
    return decrypted_data


def decrypt_file(input_filename, output_filename, private_key):

    # Open input and output files
    with open(input_filename, "rb") as infile, open(output_filename, "wb") as outfile:
        rsa_encrypted = infile.read(512)
        key = decrypt_aes_key(rsa_encrypted, private_key)
        iv = infile.read(16)
        # Create decryption cipher
        cipher = Cipher(algorithms.AES(key), modes.CBC(iv), backend=default_backend())
        decryptor = cipher.decryptor()
        # Read and decrypt in chunks
        while True:
            chunk = infile.read(4096)
            if len(chunk) == 0:
                break
            decrypted_chunk = decryptor.update(chunk)
            outfile.write(decrypted_chunk)

        # Finalize decryption
        outfile.write(decryptor.finalize())


def enc_to_ch10(encfile, key, tmpdir: str, iszstd: bool = False):
    p = Path(encfile)
    if iszstd:
        _ext = "zstd"
    else:
        _ext = "ch10"
    newfile = Path(tmpdir, f"{p.stem}.{_ext}")
    st = time.time()
    decrypt_file(encfile, newfile, key)
    try:
        decrypt_rate = os.path.getsize(encfile) / ((time.time() - st) * 1e6)
    except:
        decrypt_rate = 0.0
    print(f"Decrypted {encfile} to temp file {newfile} at {decrypt_rate:.1f}MBps")
    return newfile


def zstd_to_ch10(zstdfile, tmpdir: str) -> str:
    p = Path(zstdfile)
    newfile = Path(tmpdir, f"{p.stem}.ch10")
    st = time.time()
    with open(zstdfile, "rb") as compressed:
        with open(newfile, "wb") as decompressed:
            dctx = zstandard.ZstdDecompressor()
            dctx.copy_stream(compressed, decompressed)
    try:
        decompress_rate = os.path.getsize(zstdfile) / ((time.time() - st) * 1e6)
    except:
        decompress_rate = 0.0
    print(f"Decompressed {zstdfile} to temp file {newfile} at {decompress_rate:.1f}MBps")
    decompressed_size = os.path.getsize(newfile)
    compressed_size = os.path.getsize(zstdfile)
    print(f"Compression Ratio ={decompressed_size/compressed_size:.1f}")
    return newfile


def print_summary(channels: typing.Dict[int, Channels]):
    for id, ch in sorted(channels.items()):
        print(repr(ch))


def process_speed(channels: typing.Dict[int, Channels], start_time: float, end_time: float) -> float:
    """Return the processing speed in MBps"""
    total_data = 0.0
    for id, ch in sorted(channels.items()):
        total_data += ch.datavol
    rate = total_data * 8 / (1e6 * (end_time - start_time))
    return rate


def get_recording_rate(channels: typing.Dict[int, Channels], start_time: PTPTime, end_time: PTPTime) -> float:
    """Return the recording rate in Mbps"""
    if start_time is None or end_time is None:
        return 0.0

    total_data = 0.0
    for id, ch in sorted(channels.items()):
        total_data += ch.datavol
    time_delta = (end_time.seconds + end_time.nanoseconds / 1e9) - (start_time.seconds + start_time.nanoseconds / 1e9)
    if time_delta <= 0:
        return 0
    rate = total_data * 8 / (1e6 * time_delta)
    return rate


def ch10_to_pcap(ch10fname: str, folder: str) -> str:
    p = Path(ch10fname)
    return f"{folder}/{p.stem}.pcap"


def get_streamid_and_seq_of_inetx(buffer: bytes) -> typing.Tuple[int, int]:
    (ctrl, stream, seq) = struct.unpack_from(">III", buffer, 0x2A)
    if ctrl != 0x11000000:
        raise Exception("Packet is not inetx")
    return (stream, seq)


def validate_inetx_payload(buffer: bytes, expected_payload_len_words: int) -> bool:
    (first_word,) = struct.unpack_from(">H", buffer, 0x46)
    exp_buffer = struct.pack(
        f">{expected_payload_len_words}H",
        *list(map(lambda x: x % 65536, list(range(first_word, first_word + expected_payload_len_words)))),
    )
    return exp_buffer == buffer[0x46:]


def quick_check_inetx(buffer: bytes, expected_payload_len_words: int) -> bool:
    (first_word,) = struct.unpack_from(">H", buffer, 0x46)
    (last_word,) = struct.unpack(">H", buffer[-2:])
    expected_last_word = (first_word + expected_payload_len_words) % 65536
    return last_word == expected_last_word


def main(args):
    if args.key:
        all_files = glob.glob(os.path.join(args.folder, "*.enc"))
        dir = tempfile.mkdtemp()
    elif args.decompress:
        all_files = glob.glob(os.path.join(args.folder, "*.zst"))
        dir = tempfile.mkdtemp()
    else:
        all_files = glob.glob(os.path.join(args.folder, "*.ch10"))
        dir = None

    channels: typing.Dict[int, Channels] = {}
    roll_over = 256

    all_files.sort()
    st = time.time()
    first_ts = None
    latest_ts = None
    pkts = 0
    inetx_seq = {}
    wrapped_valid_count = 0
    for _file in all_files:
        if args.key:
            if args.decompress:
                zstd_file = enc_to_ch10(_file, args.key, dir, True)
                chfile = zstd_to_ch10(zstd_file, dir)
                os.remove(zstd_file)
            else:
                chfile = zstd_to_ch10(zstd_file, args.key, dir)
        else:
            if args.decompress:
                chfile = zstd_to_ch10(_file, dir)
            else:
                chfile = _file
        if args.unwrap:
            pf = pcap.Pcap(ch10_to_pcap(chfile, args.unwrap), mode="w")
            prec = pcap.PcapRecord()
            prec.set_current_time()

        print(f"Reading in {chfile}")
        fp = FileParser.FileParser(chfile)
        with fp as ch10file:
            for idx, _payload in enumerate(ch10file):
                pkts += 1
                pkt = ch11.Chapter11()
                try:
                    pkt.unpack(_payload)
                except Exception as e:
                    logging.error(
                        f"Failed to unpack data len={len(_payload)} as ch11. Error={e} Count={pkts}. Pkt {idx + 1} in file "
                    )

                else:
                    # Get the time to work out the data rate of the incoming packets

                    if pkt.datatype == DATA_TYPE_WRAPPED_ETHERNET:

                        if args.unwrap:
                            prec.payload = pkt.payload[16:]
                            pf.write(prec)
                        try:
                            (_sid, _seq) = get_streamid_and_seq_of_inetx(pkt.payload[16:])
                        except Exception as e:
                            logging.debug(f"Failed to unpacket wrapped packet. err={e}")
                        else:
                            if _sid in inetx_seq:
                                if inetx_seq[_sid] + 1 != _seq:
                                    logging.error(f"Unwrapped error. SID={_sid:#0X} seq={_seq} prev={inetx_seq[_sid]}")
                                else:
                                    wrapped_valid_count += 1
                            inetx_seq[_sid] = _seq
                            if CHECK_INETX_PAYLOAD:
                                if not quick_check_inetx(pkt.payload[16:], INETX_PAYLOAD_LEN_WORDS):
                                    logging.error("Inetx paylaod was corrupted")

                    if pkt.channelID == TIME_CHID:
                        time_pkt = chtime.TimeDataFormat1()
                        try:
                            time_pkt.unpack(pkt.payload)
                        except Exception as e:
                            logging.error(e)
                        else:
                            if first_ts is None:
                                first_ts = time_pkt.ptptime
                            else:
                                latest_ts = time_pkt.ptptime

                    if pkt.channelID in channels:
                        _ch = channels[pkt.channelID]
                        if (_ch.sequence + 1) % roll_over != pkt.sequence:
                            loss = pkt.sequence - ((_ch.sequence + 1) % roll_over)
                            _ch.dropcnt += loss
                            logging.error(
                                f"Dropped {loss} packets on channelID={pkt.channelID:#0X}. prev={_ch.sequence} cur={pkt.sequence}. pkt={pkts}"
                            )
                        _ch.pkt_count += 1
                        _ch.datavol += len(_payload)
                        _ch.sequence = pkt.sequence
                    else:
                        channels[pkt.channelID] = Channels(pkt.channelID, pkt.sequence, 0, 0, 0)
        if args.key or args.decompress:
            os.remove(chfile)

        print_summary(channels)
        rate = process_speed(channels, st, time.time())
        recrate = get_recording_rate(channels, first_ts, latest_ts)
        print(f"Validating at {rate:6.1f}MBps. Recording at {recrate:6.1f}Mbps")
        print(f"Validated wrapped inetx = {wrapped_valid_count}")

    print(f"---------- Result after {pkts} -----------")
    print_summary(channels)
    if args.key:
        rmtree(dir)
    print(f"Validated wrapped inetx = {wrapped_valid_count}")


if __name__ == "__main__":
    parser = create_parser()
    args = parser.parse_args()
    main(args)
    sys.exit(0)



__author__ = 'DCollins'
import sys
sys.path.append("..")

import unittest
import AcraNetwork.SimpleEthernet as SimpleEthernet


class afdxTest(unittest.TestCase):


    @unittest.skip("AFDX broken")
    def test_defaultAFDX(self):
        a = SimpleEthernet.AFDX()
        self.assertEqual(a.equipmentID,None)
        self.assertEqual(a.interfaceID,None)
        self.assertEqual(a.networkID,None)
        self.assertEqual(a.payload,None)
        self.assertEqual(a.sequencenum,None)
        self.assertEqual(a.vlink,None)



if __name__ == '__main__':
    unittest.main()



"""
.. module:: pcap
    :platform: Unix, Windows
    :synopsis: Class to pack and unpack pcap files

.. moduleauthor:: Diarmuid Collins <dcollins@curtisswright.com>

"""

__author__ = "Diarmuid Collins"
__copyright__ = "Copyright 2018"
__maintainer__ = "Diarmuid Collins"
__email__ = "dcollins@curtisswright.com"
__status__ = "Production"


import struct
import os
import time
import warnings


class PcapRecord(object):
    """
    Class that can be used to store one pcap record. A Pcap file contains one or more PcapRecords

    :type sec: int
    :type usec: int
    :type incl_len: int
    :type orig_len: int
    :type _packet: str
    """

    def __init__(self, now=False):
        """
        :param now: if True, record time is set to the current time.
        :type  now: bool
        """
        self.sec: int = 0  #: Second timestamp of the record. Epoch time
        self.usec: int = 0  #: Microsecond timestamp of the record
        self.incl_len: int = 0  #: The number of bytes captured and saved in the file
        self.orig_len: int = 0  #: The number of bytes as appeared on the network when captured
        self._payload: bytes = bytes()
        if now:
            self.set_current_time()

    # Use a property on packet so that the length is triggered on it changing
    @property
    def packet(self):
        """
        The payload within the pcap record. Payload is more accurate

        :rtype: bytes
        """
        return self._payload

    @packet.setter
    def packet(self, p: bytes) -> None:
        self._payload = p
        self.incl_len = len(p)
        self.orig_len = self.incl_len

    @property
    def payload(self):
        """
        The payload within the pcap record.

        :rtype: bytes
        """
        return self._payload

    @payload.setter
    def payload(self, p: bytes) -> None:
        self._payload = p
        self.incl_len = len(p)
        self.orig_len = self.incl_len

    def unpack(self, buf: bytes) -> None:
        """
        Unpack the pcap header. Pass in a buffer containing the header

        :type buf: bytes
        """

        if struct.calcsize(Pcap.RECORD_HEADER_FORMAT) != len(buf):
            raise ValueError("Header buffer is not the correct size to be a Pcap record header")
        (self.sec, self.usec, self.incl_len, self.orig_len) = struct.unpack(Pcap.RECORD_HEADER_FORMAT, buf)

    def pack(self) -> bytes:
        """
        Pack a PcapRecord into a buffer

        :rtype: bytes

        """
        if (
            self.sec is None
            or self.usec is None
            or self.incl_len is None
            or self.orig_len is None
            or self.packet is None
        ):
            raise ValueError("Cannot build record with undefined fields in the payload")

        return struct.pack(Pcap.RECORD_HEADER_FORMAT, self.sec, self.usec, self.incl_len, self.orig_len) + self.packet

    def setCurrentTime(self) -> bool:
        return self.set_current_time()

    def set_current_time(self):
        """
        Convienece method to set the time of the PCAP record

        :rtype: bool
        """
        currenttime = time.time()
        self.usec = int((currenttime % 1) * 1e6)
        self.sec = int(currenttime)
        return True

    def __repr__(self):
        return "LEN:{} SEC:{} USEC:{}".format(self.orig_len, self.sec, self.usec)

    def __len__(self):
        return len(self._payload)


class Pcap(object):
    """
    Create a new Pcap object with the specified filename.
    Set the mode to define read, write or append

    :param filename: The PCAP filename
    :type filename: str


    :Keyword Arguments:
        * *mode* -- r: read w: write a: append


    Pcap files look like::

        -------------- --------------- ---------------- --------------- ---------------- -------
        Global Header | Record Header | Record payload | Record Header | Record payload | .....
        -------------- --------------- ---------------- --------------- ---------------- -------

    So after opening the file, iterate through the object to read the records

    A PCAP file can be opened for reading or writing by specifying mode "r"
    or "w", or for append by specifying "a".
    
    When a PCAP file is open for writing or appending, PcapRecord objects can 
    be written to it.
    
    # Write 10 UDP records to a file
    # For simplicity use the same MAC, IP and UDP headers in all records
    >>> headers = (bytes((0x77,0x88,0x99,0xAA,0xBB,0xCC,0x66,0x55,0x44,0x33,0x22,0x11,
    ...                   0x08,0x00))
    ...           +bytes((0x45,0x00,0x00,0x36,0x77,0x77,0x40,0x00,0xff,0x11,0x8a,0xa4,
    ...                   0x12,0x34,0x56,0x78,0x99,0x88,0x77,0x66))
    ...           +bytes((0x12,0x34,0x56,0x78,0x00,0x22,0x00,0x00))
    ...           )
    >>> with Pcap("_dummy.pcap", mode='w') as p:
    ...     r = PcapRecord()
    ...     for i in range(10):
    ...         start_ch = ord('A') + i
    ...         r.payload = headers + bytes((x for x in range(start_ch,start_ch+26)))
    ...         p.write(r)
    
    When a PCAP file is open for reading, iterate through the records.
    >>> with Pcap("_dummy.pcap", mode='r') as p2:
    ...     print(f"{p2.filename} contains {p2.filesize} bytes and is open with mode '{p2.mode}'")
    ...     print(f"Network type ID {p2.network}{' (Ethernet)' if p2.network==1 else ''}")
    ...     for ix, record in enumerate(p2):
    ...         print(f"{ix} {record.orig_len} bytes: {record.payload}")
    ...
    _dummy.pcap contains 864 bytes and is open with mode 'r'
    Network type ID 1 (Ethernet)
    0 68 bytes: b'w\x88\x99\xaa\xbb\xccfUD3"\x11\x08\x00E\x00\x006ww@\x00\xff\x11\x8a\xa4\x124Vx\x99\x88wf\x124Vx\x00"\x00\x00ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    1 68 bytes: b'w\x88\x99\xaa\xbb\xccfUD3"\x11\x08\x00E\x00\x006ww@\x00\xff\x11\x8a\xa4\x124Vx\x99\x88wf\x124Vx\x00"\x00\x00BCDEFGHIJKLMNOPQRSTUVWXYZ['
    2 68 bytes: b'w\x88\x99\xaa\xbb\xccfUD3"\x11\x08\x00E\x00\x006ww@\x00\xff\x11\x8a\xa4\x124Vx\x99\x88wf\x124Vx\x00"\x00\x00CDEFGHIJKLMNOPQRSTUVWXYZ[\\'
    3 68 bytes: b'w\x88\x99\xaa\xbb\xccfUD3"\x11\x08\x00E\x00\x006ww@\x00\xff\x11\x8a\xa4\x124Vx\x99\x88wf\x124Vx\x00"\x00\x00DEFGHIJKLMNOPQRSTUVWXYZ[\\]'
    4 68 bytes: b'w\x88\x99\xaa\xbb\xccfUD3"\x11\x08\x00E\x00\x006ww@\x00\xff\x11\x8a\xa4\x124Vx\x99\x88wf\x124Vx\x00"\x00\x00EFGHIJKLMNOPQRSTUVWXYZ[\\]^'
    5 68 bytes: b'w\x88\x99\xaa\xbb\xccfUD3"\x11\x08\x00E\x00\x006ww@\x00\xff\x11\x8a\xa4\x124Vx\x99\x88wf\x124Vx\x00"\x00\x00FGHIJKLMNOPQRSTUVWXYZ[\\]^_'
    6 68 bytes: b'w\x88\x99\xaa\xbb\xccfUD3"\x11\x08\x00E\x00\x006ww@\x00\xff\x11\x8a\xa4\x124Vx\x99\x88wf\x124Vx\x00"\x00\x00GHIJKLMNOPQRSTUVWXYZ[\\]^_`'
    7 68 bytes: b'w\x88\x99\xaa\xbb\xccfUD3"\x11\x08\x00E\x00\x006ww@\x00\xff\x11\x8a\xa4\x124Vx\x99\x88wf\x124Vx\x00"\x00\x00HIJKLMNOPQRSTUVWXYZ[\\]^_`a'
    8 68 bytes: b'w\x88\x99\xaa\xbb\xccfUD3"\x11\x08\x00E\x00\x006ww@\x00\xff\x11\x8a\xa4\x124Vx\x99\x88wf\x124Vx\x00"\x00\x00IJKLMNOPQRSTUVWXYZ[\\]^_`ab'
    9 68 bytes: b'w\x88\x99\xaa\xbb\xccfUD3"\x11\x08\x00E\x00\x006ww@\x00\xff\x11\x8a\xa4\x124Vx\x99\x88wf\x124Vx\x00"\x00\x00JKLMNOPQRSTUVWXYZ[\\]^_`abc'


    The pcap can also be treated a list to select the relevant object.


    >>> # Create a pcap file
    >>> p = Pcap("_dummy.pcap", mode='w')
    >>> r = PcapRecord()
    >>> r.payload = bytes(1)
    >>> p.write(r)
    >>> p.close()
    >>> # Now open and read it
    >>> p2 = Pcap(os.path.join("_dummy.pcap"))
    >>> print(p2.network)
    1
    >>> import struct
    >>> for mypcaprecord in p2:
    ...    (firstbyte,) = struct.unpack(">B", mypcaprecord.payload)
    ...    print(firstbyte)
    0



    """

    GLOBAL_HEADER_FORMAT = "<IhhiIII"
    RECORD_HEADER_FORMAT = "<IIII"
    RECORD_HEADER_SIZE = struct.calcsize(RECORD_HEADER_FORMAT)
    GLOBAL_HEADER_SIZE = struct.calcsize(GLOBAL_HEADER_FORMAT)

    def __init__(self, filename: str, **kwargs):
        self.filename: str = filename  #: The filename of the PCAP file
        self.mode: str = kwargs.get("mode", "r")  #: The file reading mode
        self._bufferring: int = kwargs.get("buffering ", -1)  #: The file reading mode
        # Global header fields
        self.magic: int = 0xA1B2C3D4  #: The magic_number which defines the file format. Leave as is.
        self.versionmaj: int = 2  #: File format major version. Currently 2
        self.versionmin: int = 4  #: File format minor version. Currently 4
        self.zone: int = 0  #: The timezone correction in seconds. 0 = GMT
        self.sigfigs: int = 0  #: Set to 0
        self.snaplen: int = 65535  #: snapshot length. Typically unchanged
        self.network: int = 1  #: Link-layer header type. http://www.tcpdump.org/linktypes.html
        self.filesize = 0

        # rec_no is convenient if the file is not being read in a simple loop,
        # so "for ix, record in enumerate(p)" cannot easily be used. It is the 
        # number of the last record read or written. Wireshark numbers records 
        # starting at 1, so after the first record is read or written, rec_no 
        # will be 1
        self.rec_no = 0

        self.fopen = None # make deterministic if the file open fails
        try:
            self.fopen = open(filename, f"{self.mode}b", self._bufferring)
        except Exception as e:
            raise IOError(f"Failed to open {self.filename}. err={e}")

        # mode can be anything that open() will accept with 'b' added. So it
        # could be "a". But for "r" or "w", then the header must be handled.
        if self.mode == "r":
            self._read_global_header()
        elif self.mode == "w":
            self._write_global_header()

        try:
            self.filesize = os.path.getsize(filename)
        except Exception as e:
            self.filesize = 0

    # Define __enter__() and __exit__() so "with Pcap(...) as x" can be used
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def flush(self):
        return self.fopen.flush()

    def _read_global_header(self) -> bool:
        """
        This method will read the pcap global header and unpack it and propogate the relevant attributes
        This should be the first method to call on reading a pcap] file

        :rtype: bool
        """

        header = self.fopen.read(Pcap.GLOBAL_HEADER_SIZE)
        (
            self.magic,
            self.versionmaj,
            self.versionmin,
            self.zone,
            self.sigfigs,
            self.snaplen,
            self.network,
        ) = struct.unpack(Pcap.GLOBAL_HEADER_FORMAT, header)

        return True

    def _write_global_header(self):
        """
        Write the global header to a new pcap file

        :rtype: None
        """
        header = struct.pack(
            Pcap.GLOBAL_HEADER_FORMAT,
            self.magic,
            self.versionmaj,
            self.versionmin,
            self.zone,
            self.sigfigs,
            self.snaplen,
            self.network,
        )
        self.fopen.write(header)
        self.filesize += len(header)
        return True

    def write(self, pcaprecord: PcapRecord):
        """
        Write the supplied pcaprecord to the pcap file

        :param pcaprecord: The Pcap Record to write
        :type pcaprecord: PcapRecord
        """

        _pkt = pcaprecord.pack()
        self.fopen.write(_pkt)
        self.rec_no += 1
        self.filesize += len(_pkt)

    def close(self):
        """
        Close the current pcap file

        :rtype: None
        """
        self.fopen.close()

    def __iter__(self):
        return self

    def next(self):
        # read the pcap header to a new object
        pcaprecord = PcapRecord()
        try:
            pcaprecord.unpack(self.fopen.read(Pcap.RECORD_HEADER_SIZE))
        except:
            raise StopIteration

        try:
            pcaprecord.packet = self.fopen.read(pcaprecord.incl_len)
        except:
            raise StopIteration
        else:
            self.rec_no += 1
            return pcaprecord

    __next__ = next

    def __getitem__(self, item):
        self.fopen.seek(Pcap.GLOBAL_HEADER_SIZE)
        for idx, rec in enumerate(self):
            if idx == item:
                return rec



__author__ = "DCollins"

import sys

# sys.path.append("../AcraNetwork")
import os

import unittest
import AcraNetwork.Pcap as pcap
import AcraNetwork.SimpleEthernet as SimpleEthernet
import AcraNetwork.iNetX as inetx
import AcraNetwork.MPEGTS as MPEGTS
import base64
import AcraNetwork.MPEG.PMT as MPEGPMT
import AcraNetwork.MPEG.PES as pes
import struct
import datetime
import logging
import random

THIS_DIR = os.path.dirname(os.path.abspath(__file__))


mpeg_ts_repr = """PID=0X0 PUSI=True TSC=Not Scrambled Adaption=Payload Only
PID=0X1000 PUSI=True TSC=Not Scrambled Adaption=Payload Only
PID=0X100 PUSI=True TSC=Not Scrambled Adaption=Adaption and Payload
 Adaption=Discontunity=False, random=True Elementary Stream Indicator=False PCR=True OPCR=False Splicing Point Flag=False Transport Private Data=False, Adaption Extension=False
PID=0X100 PUSI=False TSC=Not Scrambled Adaption=Payload Only
PID=0X100 PUSI=False TSC=Not Scrambled Adaption=Payload Only
PID=0X100 PUSI=False TSC=Not Scrambled Adaption=Payload Only
PID=0X100 PUSI=False TSC=Not Scrambled Adaption=Payload Only
"""


class MPEGTSBasicTest(unittest.TestCase):

    ######################
    # Read a complete pcap file
    ######################

    def test_readFirstMPEGTS(self):
        """
        Very simple test that reads a pcap file with mpegts packets.
        Takes the first packet in there and decoms the mpegts blocks
        Verifies each block in that first packet
        """
        p = pcap.Pcap(os.path.join(THIS_DIR, "mpegts_input.pcap"))
        mypcaprecord = p[0]

        ethpacket = SimpleEthernet.Ethernet()  # Create an Ethernet object
        ethpacket.unpack(mypcaprecord.packet)  # Unpack the pcap record into the eth object
        ippacket = SimpleEthernet.IP()  # Create an IP packet
        ippacket.unpack(ethpacket.payload)  # Unpack the ethernet _payload into the IP packet
        udppacket = SimpleEthernet.UDP()  # Create a UDP packet
        udppacket.unpack(ippacket.payload)  # Unpack the IP _payload into the UDP packet
        inetxpacket = inetx.iNetX()  # Create an iNetx object
        inetxpacket.unpack(udppacket.payload)  # Unpack the UDP _payload into this iNetX object

        mpegts = MPEGTS.MPEGTS()
        mpegts.unpack(inetxpacket.payload)
        self.assertEqual(len(mpegts), 7)
        for packet_index in list(range(7)):
            if packet_index == 0:
                self.assertEqual(mpegts.blocks[packet_index].pid, 0)
                self.assertEqual(mpegts.blocks[packet_index].sync, 0x47)
            elif packet_index == 1:
                self.assertEqual(mpegts.blocks[packet_index].pid, 4096)
                self.assertEqual(mpegts.blocks[packet_index].sync, 0x47)
            else:
                self.assertEqual(mpegts.blocks[packet_index].pid, 256)
                self.assertEqual(mpegts.blocks[packet_index].sync, 0x47)
        p.close()

    def test_pack_unpack_adaption(self):
        ad = MPEGTS.MPEGAdaption()
        adex = MPEGTS.MPEGAdaptionExtension()
        ad.adaption_extension = adex
        packed = ad.pack()
        ad2 = MPEGTS.MPEGAdaption()
        ad2.unpack(packed)
        self.assertEqual(ad2, ad)

    def test_pack_unpack(self):
        """
        Very simple test that reads a pcap file with mpegts packets.
        Takes the first packet in there and decoms the mpegts blocks
        Verifies each block in that first packet
        """
        p = pcap.Pcap(os.path.join(THIS_DIR, "mpegts_input.pcap"))
        mypcaprecord = p[0]
        self.maxDiff = None
        mpegts = MPEGTS.MPEGTS()
        payload = mypcaprecord.payload[0x46:]
        mpegts.unpack(payload)
        self.assertEqual(mpeg_ts_repr, repr(mpegts))
        self.assertEqual(len(mpegts), 7)
        p.close()
        for block in mpegts:
            if block.adaption_field is not None:
                pass
        packed = mpegts.pack()
        self.assertEqual(packed, payload)

        mpegts2 = MPEGTS.MPEGTS()
        mpegts2.unpack(packed)
        self.assertEqual(mpegts, mpegts2)

    def test_readAllMPEGTS(self):
        """
        Reads the same mpeg ts file as previously but reads all the data
        in the file and checks for any continuity errors
        :return:
        """
        p = pcap.Pcap(os.path.join(THIS_DIR, "mpegts_input.pcap"))
        for mypcaprecord in p:

            ethpacket = SimpleEthernet.Ethernet()  # Create an Ethernet object
            ethpacket.unpack(mypcaprecord.packet)  # Unpack the pcap record into the eth object
            ippacket = SimpleEthernet.IP()  # Create an IP packet
            ippacket.unpack(ethpacket.payload)  # Unpack the ethernet _payload into the IP packet
            udppacket = SimpleEthernet.UDP()  # Create a UDP packet
            udppacket.unpack(ippacket.payload)  # Unpack the IP _payload into the UDP packet
            inetxpacket = inetx.iNetX()  # Create an iNetx object
            inetxpacket.unpack(udppacket.payload)  # Unpack the UDP _payload into this iNetX object
            mpegts = MPEGTS.MPEGTS()
            mpegts.unpack(inetxpacket.payload)
            self.assertEqual(len(mpegts), 7)

        p.close()

    @unittest.skip("Broken")
    def test_stanag(self):
        ts_file = open(os.path.join(THIS_DIR, "stanag_sample.ts"), mode="rb")
        h264_data = MPEGTS.H264()
        self.assertTrue(h264_data.unpack(ts_file.read()))
        ts_file.close()
        self.assertEqual(len(h264_data.nals), 258)
        nal_counts = {}
        for nal in h264_data.nals:
            if not nal.type in nal_counts:
                nal_counts[nal.type] = 1
            else:
                nal_counts[nal.type] += 1
        self.assertEqual(nal_counts[0], 35)
        self.assertEqual(nal_counts[6], 70)

    @unittest.skip("Missing input ts")
    def test_capture(self):
        ts_f = open(os.path.join(THIS_DIR, "capture_0XDC1.ts"), mode="rb")
        pids = {}
        PMT_PID = 256
        while True:
            ts_block = ts_f.read(188)
            if not ts_block:
                break
            mpeg_pkt = MPEGTS.MPEGPacket()
            try:
                mpeg_pkt.unpack(ts_block)
            except Exception as e:
                logging.error(e)
                # self.assertTrue(False)
            else:
                if mpeg_pkt.pid not in pids:
                    pids[mpeg_pkt.pid] = 0
                pids[mpeg_pkt.pid] += 1

                if mpeg_pkt.pid == PMT_PID:
                    pmt_pkt = MPEGPMT.MPEGPacketPMT()
                    try:
                        pmt_pkt.unpack(ts_block)
                    except Exception as e:
                        logging.error(f"Failed to unpack offset {ts_f.tell():#0X} as PMT")
                        sys.exit(1)
                    else:
                        # print(f"PMT={repr(pmt_pkt)}")
                        pass

        ts_f.close()
        for pid, count in sorted(pids.items()):
            logging.info(f"PID={pid:#0X} Count={count}")


# Take from Wireshar - > cop
pmt_payload = base64.b64decode(
    "R0EAGQACsDgAAcEAAOEB8AAb4QLwAAPhA/AAFeEE8BwFBEtMVkEmCQEA/0tMVkEADycJwQAAwIAAwQAAX+WEI/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////8="
)

# Two descriptor tags + twp streams
pmt2_payload = base64.b64decode(
    "R0EAEQACsCMAAcEAAPAA8AwFBEhETVaIBA///Pwb8QDwAA/xEPAAn17xO/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////8="
)


class MPEGPMTTest(unittest.TestCase):
    def test_pmt_unpack(self):
        pmt = MPEGPMT.MPEGPacketPMT()
        pmt.unpack(pmt_payload)

        pmt2 = MPEGPMT.MPEGPacketPMT()
        pmt2.unpack(pmt.pack())
        self.assertEqual(pmt, pmt2)
        self.assertEqual(pmt_payload, pmt2.pack())

    def test_pmt2_unpack(self):
        pmt = MPEGPMT.MPEGPacketPMT()
        pmt.unpack(pmt2_payload)

    def test_pmt_random(self):
        pmt = MPEGPMT.MPEGPacketPMT()
        pmt.sync = 0x47
        pmt.adaption_ctrl = MPEGTS.ADAPTION_PAYLOAD_ONLY
        pmt.pid = 0x100
        pmt.continuitycounter = 7
        pmt.last_section = 0xE
        pmt.section = 0xD
        pmt.program_number = 0xDEAD
        t = MPEGPMT.DescriptorTag()
        t.data = struct.pack(">5H", *(list(range(5))))
        t.tag = 0xA
        s = MPEGPMT.PMTStream()
        pmt.descriptor_tags = [t]
        s.elementary_pid = 0x1FEF
        pmt.streams = [s]
        _packed = pmt.pack()

        pmt2 = MPEGPMT.MPEGPacketPMT()
        self.assertTrue(pmt2.unpack(_packed))
        self.assertEqual(pmt, pmt2)


pes_packet = base64.b64decode(
    "R0EEP4UA////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////AAAB/AAsgYAFIQQD/tEAD98AHwYOKzQCCwEBDgEDAQEAAAAOAggABg/Gi5FmYwECH1M="
)
stanag_packet = base64.b64decode(
    "R0EEMYUA////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////AAAB/AAsgYAFAQQF2TcAAd8AHwYOKzQCCwEBDgEDAQEAAAAOAggABhTIE/mO6QECDwg="
)


class MPEG_PES(unittest.TestCase):
    def test_pes_unpack(self):
        self.assertEqual(188, len(pes_packet))  # Verify the length
        p = pes.PES()
        p.unpack(pes_packet)  # unpack as PES
        self.assertEqual(50, len(p.payload))
        # print(repr(p))

        self.assertEqual(36, len(p.pesdata))  # Verify the pes payload
        self.assertEqual(pes_packet, p.pack())  # Verify that pack turns it back exactly into the same packet
        p2 = pes.PES()
        p2.unpack(p.pack())  # Unpack it again
        self.assertEqual(p, p2)  # Compre packets
        self.assertEqual(p2.pack(), pes_packet)  # Verify packing

    def test_stanag_unpack(self):
        p = pes.STANAG4609()
        p.unpack(pes_packet)
        # print(repr(p))
        self.assertEqual(36, len(p.pesdata))
        self.assertEqual(len(p.pack()), 188)
        self.assertEqual(p.pack(), pes_packet)
        # print(repr(p.time))

    def test_stanag2_unpack(self):
        p = pes.STANAG4609()
        p.unpack(stanag_packet)
        # print(repr(p))
        self.assertEqual(36, len(p.pesdata))
        self.assertEqual(len(p.pack()), 188)
        self.assertEqual(p.pack(), stanag_packet)
        # print(repr(p.time))

    def test_stanag_create(self):
        ref = pes.STANAG4609()
        ref.unpack(pes_packet)
        # Build a packet from scratch that matches the refernce packet
        p = pes.STANAG4609()
        p.pid = pes.STANAG4609_PID
        p.adaption_ctrl = MPEGTS.ADAPTION_PAYLOAD_AND_ADAPTION
        p.adaption_field = MPEGTS.MPEGAdaption()
        p.adaption_field.length = 133
        p.time = datetime.datetime(2024, 1, 25, 15, 7, 59, 767139, tzinfo=datetime.timezone.utc)
        p.header_data = pes.ts_to_buf(187.14)
        p.extension_w1 = pes.PES_EXTENSION_W1
        p.extension_w2 = pes.PES_EXTENSION_W2
        p.streamid = 0xFC
        p.stanag_counter = 15
        p.continuitycounter = 15
        p.pack()
        self.assertEqual(p, ref)
        self.assertEqual(p.pack(), pes_packet)

    def test_stanag_pts(self):
        ts = 0x210403FED1
        exp = 187.14
        self.assertEqual(pes.pts_to_ts(ts), exp)
        _conv = pes.ts_to_pts(exp)
        # print(f"{_conv:#0X}")
        self.assertEqual(_conv, ts)


corrupt_vid_ = base64.b64decode(
    "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAEdAEREABBYXGBkaGxwdHh8gISIjJCUmJygpKissLS4vMDEyMzQ1Njc4OTo7PD0+P0BBQkNERUZHSElKS0xNTk9QUVJTVFVWV1hZWltcXV5fYGFiY2RlZmdoaWprbG1ub3BxcnN0dXZ3eHl6e3x9fn+AgYKDhIWGh4iJiouMjY6PkJGSk5SVlpeYmZqbnJ2en6ChoqOkpaanqKmqq6ytrq+wsbKztLW2t7i5uru8vb6/wMHCw8TFxsdeSR2TR0AQEQAFyMnKy8zNzs/Q0dLT1NXW19jZ2tvc3d7f4OHi4+Tl5ufo6err7O3u7/Dx8vP09fb3+Pn6+/z9/v8AAQIDBAUGBwgJCgsMDQ4PEBESExQVFhcYGRobHB0eHyAhIiMkJSYnKCkqKywtLi8wMTIzNDU2Nzg5Ojs8PT4/QEFCQ0RFRkdISUpLTE1OT1BRUlNUVVZXWFlaW1xdXl9gYWJjZGVmZ2hpamtsbW5vcHFyc3R1dnd4eRnP0dRHH/8QAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAEcf/xAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAARx//EAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAABHH/8QAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAEcf/xAAAAAAAAAAAAAAAAAAAAAA"
)


def gen_random_adaption() -> MPEGTS.MPEGAdaption:
    """Generate a random Adaption block

    Returns:
        mpegts.MPEGAdaption: _description_
    """
    adaption = MPEGTS.MPEGAdaption()
    adaption_ext = MPEGTS.MPEGAdaptionExtension()
    adaption_ext.ltw = struct.pack(">H", 0x1)
    adaption.adaption_extension = adaption_ext
    adaption.pcr = struct.pack(">HI", 0xDC, 0xA)
    private_data_len = 8  # random.randint(4, 10)
    adaption.private_data = struct.pack(f">{private_data_len}B", *range(private_data_len))
    return adaption


AUD_WIRESHARK = base64.b64decode(
    "R1EQNAAAAAHAALGBwAohAYttlxEBi22X//lMgBSf/CEWzYAAAJGxQZJoFN2gK0IqYD978ffG6HO42VQXy00TEmkAPFSnXfBXIe2Iq9aWg0UiXK5OXAUO1U3QmVYV5t3JdDMqertkQROtoUFHQrF3eSojHSC6qoVwBSICg5qfL+OhrglxPOQMYCYb2uchqrcVY9Jztkq9mJA5SZbxyBwLAVqbVR64t7xX3eRlyO2UTpnssSgU0sUFHQqfF34="
)
AUD_PAYLOAD = base64.b64decode(
    "//lMgBSf/CEWzYAAAJGxQZJoFN2gK0IqYD978ffG6HO42VQXy00TEmkAPFSnXfBXIe2Iq9aWg0UiXK5OXAUO1U3QmVYV5t3JdDMqertkQROtoUFHQrF3eSojHSC6qoVwBSICg5qfL+OhrglxPOQMYCYb2uchqrcVY9Jztkq9mJA5SZbxyBwLAVqbVR64t7xX3eRlyO2UTpnssSgU0sUFHQqfF34="
)

AUD_WIRESHARK2 = base64.b64decode(
    "R1EQOC4A////////////////////////////////////////////////////////////AAABwACDgcAKIQGNjXkRAY2Nef/5TIAO3/whFs/V/++SoqIYaLUKxbLXVWoq5AYet7svXAIhmHcsb64qz2LtaB5KVl4xYGSCUtowuItllZjKsAQJWCYTNB99yj5UGUVa3Ar4V1XHaw4aUzFjjQx3k2UDECBEsJVUMLPvWs7VhHRgCLxSkKD77lw="
)

PAT_PKT = base64.b64decode(
    "R0AAHAAAsA0ACMEAAAAB4QBn1F+J//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////8="
)
PMT_PKT = base64.b64decode(
    "R0EAHAACsEQAAcEAAPAA8AwFBEhETVaIBA///Pwb8QDwAATxEPAAFeEE8BwFBEtMVkEmCQEA/0tMVkEADycJwQAAwIAAwQAA7IayY/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////8="
)

AUD_PAYLOAD2 = base64.b64decode(
    "//lMgA7f/CEWz9X/75KiohhotQrFstdVairkBh63uy9cAiGYdyxvrirPYu1oHkpWXjFgZIJS2jC4i2WVmMqwBAlYJhM0H33KPlQZRVrcCvhXVcdrDhpTMWONDHeTZQMQIESwlVQws+9aztWEdGAIvFKQoPvuXA=="
)


class MPEGAdaptionFullTest(unittest.TestCase):

    def test_fromsim(self):
        f = open(THIS_DIR + "/mpegs_ch0.ts", mode="br")
        f.seek(752)
        while True:
            print(f"Reading from offset {f.tell()}")
            buf = f.read(188)
            if len(buf) == 0:
                break
            mp = MPEGTS.MPEGPacket()
            try:
                mp.unpack(buf)
            except Exception as e:
                print(e)
                self.assertTrue(False)

    def test_pack_and_unpack(self):
        for _iter in range(1 - 0):
            mpeg_pkt = MPEGTS.MPEGPacket()
            mpeg_pkt.sync = 0x47
            mpeg_pkt.transport_priority = random.choice([0, 1])  # 1 in 10
            mpeg_pkt.tei = not random.randint(0, 100)  # 1 in 100
            mpeg_pkt.pid = random.randint(1, 5000)  # Select a PID
            mpeg_pkt.pusi = random.choice([True, False])
            mpeg_pkt.continuitycounter = random.randint(1, 6)
            mpeg_pkt.adaption_ctrl = random.choice([1, 2])
            adaption_field_len = 0
            if mpeg_pkt.adaption_ctrl == 0x2 or mpeg_pkt.adaption_ctrl == 0x3:
                mpeg_pkt.adaption_field = gen_random_adaption()
                adaption_field_len = len(mpeg_pkt.adaption_field.pack())
            payload_len = 188 - 4 - adaption_field_len
            if mpeg_pkt.adaption_ctrl != MPEGTS.ADAPTION_ADAPTION_ONLY:
                mpeg_pkt.payload = struct.pack(f"{payload_len}B", *range(payload_len))

            buf = mpeg_pkt.pack()

            mp = MPEGTS.MPEGPacket()
            self.assertIsNone(mp.unpack(buf))
            self.assertEqual(mp, mpeg_pkt)

    def test_unpack_audio_adaption(self):
        pkt = pes.PES()
        pkt.unpack(AUD_WIRESHARK)
        self.assertEqual(len(pkt.header_data), 10)
        pts = pes.buf_to_ts(pkt.header_data[:5])
        dts = pes.buf_to_ts(pkt.header_data[5:])
        self.assertAlmostEqual(pts, 71.881366666)
        self.assertAlmostEqual(dts, 71.881366666)

        pkt2 = pes.PES()
        pkt2.tei = False
        pkt2.pusi = True
        pkt2.transport_priority = 0
        pkt2.pid = 0x1110
        pkt2.continuitycounter = 4
        pkt2.tsc = 0
        pkt2.adaption_ctrl = MPEGTS.ADAPTION_PAYLOAD_AND_ADAPTION
        pkt2.streamid = 0xC0
        pkt2.extension_w1 = 0x81
        pkt2.extension_w2 = 0xC0
        pkt2.header_data = pes.ts_to_buf(71.881366666) + pes.ts_to_buf(71.881366666)
        pkt2.pesdata = AUD_PAYLOAD

        buf = pkt2.pack()
        # self.assertEqual(buf, AUD_WIRESHARK)
        f = open(THIS_DIR + "/audio_gen.ts", mode="bw")
        f.write(PAT_PKT)
        f.write(PMT_PKT)
        # f.write(AUD_WIRESHARK)
        f.write(buf)
        f.close()

    def test_unpack_audio_adaption2(self):
        pkt = pes.PES()
        pkt.unpack(AUD_WIRESHARK2)
        self.assertEqual(len(pkt.header_data), 10)
        pts = pes.buf_to_ts(pkt.header_data[:5])
        dts = pes.buf_to_ts(pkt.header_data[5:])
        self.assertAlmostEqual(pts, 72.290800000)
        self.assertAlmostEqual(dts, 72.290800000)

        pkt2 = pes.PES()
        pkt2.tei = False
        pkt2.pusi = True
        pkt2.transport_priority = 0
        pkt2.pid = 0x1110
        pkt2.continuitycounter = 8
        pkt2.tsc = 0
        pkt2.adaption_ctrl = MPEGTS.ADAPTION_PAYLOAD_AND_ADAPTION
        pkt2.adaption_field = MPEGTS.MPEGAdaption()
        pkt2.adaption_field.length = 46

        pkt2.streamid = 0xC0
        pkt2.extension_w1 = 0x81
        pkt2.extension_w2 = 0xC0
        pkt2.header_data = pes.ts_to_buf(72.290800000) + pes.ts_to_buf(72.290800000)
        pkt2.pesdata = AUD_PAYLOAD2

        buf = pkt2.pack()
        f = open(THIS_DIR + "/audio_gen2.ts", mode="bw")
        f.write(PAT_PKT)
        f.write(PMT_PKT)
        # f.write(AUD_WIRESHARK)
        f.write(buf)
        f.close()

    def test_random_pes(self):

        for _i in range(100):
            pkt2 = pes.PES()
            pkt2.tei = False
            pkt2.pusi = True
            pkt2.transport_priority = 0
            pkt2.pid = 1
            pkt2.continuitycounter = 2
            pkt2.tsc = 0
            pkt2.adaption_ctrl = MPEGTS.ADAPTION_PAYLOAD_AND_ADAPTION
            adaption = random.choice([0, 12, 25, 44])
            if adaption != 0:
                pkt2.adaption_field = MPEGTS.MPEGAdaption()
                pkt2.adaption_field.length = adaption
            pkt2.streamid = 0xC0
            pkt2.extension_w1 = 0x81
            pkt2.extension_w2 = 0xC0
            pkt2.header_data = pes.ts_to_buf(44) + pes.ts_to_buf(44)
            pkt2.pesdata = struct.pack(f">{164 - adaption}B", *range(164 - adaption))
            if len(pkt2.pack()) != 188:
                pass
            self.assertEqual(len(pkt2.pack()), 188)


class MPEGAdaptionExtensionTestcase(unittest.TestCase):
    def test_basic(self):
        for _i in range(100):
            _e = MPEGTS.MPEGAdaptionExtension()
            _e.ltw = random.choice([bytes(), b"\x00\x00"])
            _e.piecewise = random.choice([bytes(), b"\x00\x00\x00"])
            _e.seamless_splice = random.choice([bytes(), b"\x00\x00\x00\x00\x00"])

            _d = MPEGTS.MPEGAdaptionExtension()
            _d.unpack(_e.pack())
            self.assertEqual(_d, _e)


class MPEGAdaptionTestcase(unittest.TestCase):
    def test_basic(self):
        for _i in range(100):
            _e = MPEGTS.MPEGAdaption()
            adaption_ext = MPEGTS.MPEGAdaptionExtension()
            adaption_ext.ltw = struct.pack(">H", 0x1)
            _e.adaption_extension = adaption_ext
            _e.pcr = random.choice([struct.pack(">HI", 0xDC, 0xA), bytes()])
            private_data_len = random.randint(4, 10)
            _e.private_data = random.choice([struct.pack(f">{private_data_len}B", *range(private_data_len)), bytes()])

            _d = MPEGTS.MPEGAdaption()
            _b = _e.pack()
            _d.unpack(_b)
            self.assertEqual(_e, _d)


class CorruptVid(unittest.TestCase):
    def test_basic(self):
        f = open(THIS_DIR + "/audio_pkt.ts", mode="bw")
        f.write(PAT_PKT)
        f.write(PMT_PKT)
        for _iter in range(1):
            mpeg_pkt = MPEGTS.MPEGPacket()
            mpeg_pkt.sync = 0x47
            mpeg_pkt.transport_priority = random.choice([0, 1])  # 1 in 10
            mpeg_pkt.tei = not random.randint(0, 100)  # 1 in 100
            mpeg_pkt.pid = 0x1110  # Select a PID
            mpeg_pkt.pusi = False
            mpeg_pkt.continuitycounter = random.randint(1, 6)
            mpeg_pkt.adaption_ctrl = random.choice([1, 2])
            adaption_field_len = 0
            if mpeg_pkt.adaption_ctrl == 0x2 or mpeg_pkt.adaption_ctrl == 0x3:
                mpeg_pkt.adaption_field = gen_random_adaption()
                adaption_field_len = len(mpeg_pkt.adaption_field.pack())
            payload_len = 188 - 4 - adaption_field_len
            if mpeg_pkt.adaption_ctrl != MPEGTS.ADAPTION_ADAPTION_ONLY:
                mpeg_pkt.payload = struct.pack(f"{payload_len}B", *range(payload_len))

            buf = mpeg_pkt.pack()
            f.write(buf)
        f.close()


if "unittest.util" in __import__("sys").modules:
    # Show full diff in self.assertEqual.
    __import__("sys").modules["unittest.util"]._MAX_LENGTH = 999999999

if __name__ == "__main__":
    unittest.main()



import sys
sys.path.append("..")

import unittest
from AcraNetwork.ptptime import ptptime, timefromsbi


class PTPTestCase(unittest.TestCase):

    def test_something(self):

        t = ptptime(2000,1,2,0,0,0,0)
        #print(t.irigtime())

        t2 = ptptime(year=1970, month=1, day=1, hour=23, minute=1, second=20, microsecond=100, nanosecond=1)
        #print t2
        (hi, lo ,mu) = t2.irigtime()
        #print "hi={:#0X} lo={:#0X} mu={:#0X}".format(hi, lo, mu)
        #print "hr={:#0X} min={:#0X} lo={:#0X} mu={:#0X}".format(hi>>7, hi&0x7F, lo, mu)

    def test_irig(self):
        s = 0x123456
        t = timefromsbi(s)
        #print(t.microsecond)

if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
import struct


class Analog(object):
    """
    Class to handle Chapter11 analog packets
    https://www.irig106.org/docs/106-22/chapter11.pdf
    11.2.5.2

    Unpack the chapter10 payload into this format
    """

    def __init__(self):
        self.channel_specific_word: int = 0
        self.data: bytes = bytes()

    def unpack(self, buffer: bytes):
        """
        Convert a string buffer into a Analog
        :type buffer: bytes
        :rtype: bool
        """

        (self.channel_specific_word,) = struct.unpack_from("<I", buffer, 0)
        self.data = buffer[4:]

        return True

    def pack(self):
        return struct.pack("<I", self.channel_specific_word) + self.data

    def __repr__(self):
        _rstr = "Chapter10 Analog Channel Specific Word ={:#0X}\n".format(self.channel_specific_word)
        return _rstr

    def __eq__(self, other: Analog):
        """

        :type other: Analog
        :return:
        """
        if not isinstance(other, Analog):
            return False

        if self.channel_specific_word != other.channel_specific_word:
            return False

        if len(self.data) != len(other.data):
            return False

        return True



import struct
import datetime


SEI_UNREG_DATA = 5


class STANAG4609_SEI(object):
    """
    Handle the SEI NAL and more specifically this will handle SEIs defined in 3.14.3.5 of the STANAG standard
    http://www.gwg.nga.mil/misb/docs/nato_docs/STANAG_4609_Ed3.pdf
    """

    def __init__(self):
        self.payloadtype = None
        self.payloadsize = None
        self.unregdata = False
        self.status = None
        self.seconds = None
        self.microseconds = None
        self.nanoseconds = None
        self.time = None
        self.stanag = False

    def unpack(self, buf):
        """
        Unpack the NAL _payload as an STANAG4609_SEI

        :param buf: The buffer to unpack into an STANAG4609_SEI
        :type buf: str
        :rtype: bool
        """

        (self.payloadtype, self.payloadsize) = struct.unpack(">BB", buf[0:2])
        if self.payloadtype == SEI_UNREG_DATA:
            self.unregdata = True
            (
                sig1,
                sig2,
                self.status,
                ms1,
                _fix1,
                ms2,
                _fix2,
                ms3,
                _fix3,
                ms4,
            ) = struct.unpack_from(">QQBHBHBHBH", buf[2:])
            # combine the time fields (cf  http://www.gwg.nga.mil/misb/docs/nato_docs/STANAG_4609_Ed3.pdf 3.14.3.4 )
            # Verify the signature and if it's good then convert to a time
            if (
                sig1 == 0x4D4953506D696372
                and sig2 == 0x6F73656374696D65
                and _fix1 == 0xFF
                and _fix2 == 0xFF
                and _fix3 == 0xFF
            ):
                useconds = (ms1 << 48) + (ms2 << 32) + (ms3 << 16) + ms4
                self.seconds = float(useconds) / 1.0e6
                self.nanoseconds = (ms3 << 16) + ms4
                self.time = datetime.datetime.fromtimestamp(self.seconds)
                self.stanag = True



from __future__ import annotations
import struct
from AcraNetwork.IRIG106.Chapter11 import TS_CH4, TS_IEEE1558, RTCTime, PTPTime
import typing


class MILSTD1553Message(object):
    """
    The Chapter 10 standard defines specific payload formats for different data. This class handles MIL1553 packets

    :type ptptimeseconds: int
    :type ptptimenanoseconds: int
    :type subchannel: int
    :type parity_error: bool
    :type datalength: int
    :type payload: str
    """

    def __init__(self, ipts_source: int = TS_CH4):
        if ipts_source == TS_CH4:
            self.ipts = RTCTime()
        elif ipts_source == TS_IEEE1558:
            self.ipts = PTPTime()
        elif ipts_source is None:
            raise Exception("Time stamp is not option for MIL-STD-1553")
        self.blockstatus: int = 0
        self.gaptimes: int = 0
        self.length: int = 0
        self.message: bytes = bytes()

    def pack(self):
        """
        Pack the MIL-STD-1553 message object into a binary buffer

        :rtype: bytes
        """
        ch_spec_word = self.ipts.pack()
        self.length = len(self.message)
        intra_packet_data_header = struct.pack("<HHH", self.blockstatus, self.gaptimes, self.length)

        return ch_spec_word + intra_packet_data_header + self.message

    def unpack(self, mybuffer: bytes):
        """
        Unpack a string buffer into an MIL-STD-1553 data packet object. Returns the buffer that was consumed

        :param mybuffer: A string buffer representing an UART data  packet
        :type mybuffer: str
        :rtype: int
        """
        offset = 0
        # bytes = struct.unpack_from(">8B", mybuffer)
        self.ipts.unpack(mybuffer[:8])
        offset += 8
        (self.blockstatus, self.gaptimes, self.length) = struct.unpack_from("<HHH", mybuffer, offset)
        offset += 6
        self.message = mybuffer[offset : offset + self.length]
        offset += self.length

        return offset

    def __eq__(self, other: MILSTD1553Message):
        if not isinstance(other, MILSTD1553Message):
            return False

        _match_att = ["ipts", "blockstatus", "gaptimes", "length", "message"]

        for attr in _match_att:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    def __repr__(self):
        return "MILSTD1553Message: Time={}  BlockStatus={} GapTimes={} Length={}".format(
            self.ipts, self.blockstatus, self.gaptimes, self.length
        )


class MILSTD1553DataPacket(object):
    """
    Data Packet Format. Contains a list of MIML-STD-1553 Data Words

    :type msgcount: int
    :type uartwords: list[MILSTD1553Message]


    >>> from AcraNetwork.IRIG106.Chapter11 import Chapter11
    >>> c = Chapter11()
    >>> m = MILSTD1553DataPacket()
    >>> milmessage = MILSTD1553Message()
    >>> milmessage.message = bytes(2)
    >>> m.append(milmessage)
    >>> c.payload = m.pack()
    >>> ch10_buffer = c.pack()
    >>> # Now reverse the process
    >>> c2 = Chapter11()
    >>> c2.unpack(ch10_buffer)
    True
    >>> m2 =  MILSTD1553DataPacket()
    >>> m2.unpack(c2.payload)
    True
    >>> print(m2)
    MILSTD1553DataPacket: MessageCount=1 [MILSTD1553Message: Time=RTC: count=0  BlockStatus=0 GapTimes=0 Length=2]
      MILSTD1553Message: Time=RTC: count=0  BlockStatus=0 GapTimes=0 Length=2
    <BLANKLINE>
    >>> print(m2.msgcount)
    1

    """

    def __init__(self, ipts_source=TS_CH4):
        self.messages: typing.List[MILSTD1553Message] = []  #: List of :class:`MILSTD1553Message`
        self.msgcount: int = 0
        self.ttb: int = 0
        self._ipts_source = ipts_source

    def pack(self):
        """
        Pack the MIL-STD-1553 data packet object into a binary buffer

        :rtype: str
        """
        # Some checks
        if len(self) == 0:
            raise Exception("No MILSTD1553Messages defined")

        msg_buf = bytes()
        for dw in self:
            msg_buf += dw.pack()

        csw = struct.pack("<I", (self.ttb << 30) + len(self))

        return csw + msg_buf

    def unpack(self, mybuffer):
        """
        Unpack a string buffer into an UART data packet object

        :param buffer: A string buffer representing an UART data  packet
        :type buffer: str
        :rtype: None
        """

        (ch_spec_word,) = struct.unpack_from("<I", mybuffer)
        self.msgcount = ch_spec_word & 0xFFFFFF
        self.ttb = (ch_spec_word >> 30) & 0x3

        offset = 4
        while offset + 14 < len(mybuffer):  # Should have at least the timestamp
            m = MILSTD1553Message(self._ipts_source)
            offset += m.unpack(mybuffer[offset:])
            self.messages.append(m)

        return True

    def append(self, message: MILSTD1553Message):
        """
        Add a message to the data packet

        :type message: MILSTD1553Message
        :rtype: bool
        """
        if not isinstance(message, MILSTD1553Message):
            raise Exception("Can only append MILSTD1553Message")
        self.msgcount += 1
        return self.messages.append(message)

    def __eq__(self, other):
        if not isinstance(other, MILSTD1553DataPacket):
            return False

        _match_att = ["messages", "msgcount"]

        for attr in _match_att:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    def __repr__(self):
        ret_str = "MILSTD1553DataPacket: MessageCount={} {}\n".format(self.msgcount, self.messages)

        for a in self:
            ret_str += "  {}\n".format(repr(a))

        return ret_str

    def __iter__(self):
        self._index = 0
        return self

    def next(self) -> MILSTD1553Message:
        if self._index < len(self.messages):
            _dw = self.messages[self._index]
            self._index += 1
            return _dw
        else:
            raise StopIteration

    __next__ = next

    def __len__(self):
        return len(self.messages)

    def __getitem__(self, key):
        return self.messages[key]



import typing
import struct
from AcraNetwork.MPEG.STANAG4609 import STANAG4609_SEI
import sys

NAL_HEADER = 0x00000001
NAL_HEADER_LEN = 4
NAL_TYPES = {
    "Unspecified": 0,
    "Coded non-IDR": 1,
    "Coded partition A": 2,
    "Coded partition B": 3,
    "Coded partition C": 4,
    "Coded IDR": 5,
    "SEI": 6,
    "SPS": 7,
    "PPS": 8,
    "AUD": 9,
    "EOSeq": 10,
    "EOStream": 11,
    "Filler": 12,
    "SES": 13,
    "Prefix NAL": 14,
    "SSPS": 15,
    "Reserved": 16,
}
# Invert it to go from integer to more useful name
NAL_TYPES_INV = {v: k for k, v in list(NAL_TYPES.items())}
PY3 = sys.version_info > (3,)


class H264(object):
    """
    This class will handle H.264 _payload. It can convert a buffer of bytes into an array
    of NALs(https://en.wikipedia.org/wiki/Network_Abstraction_Layer)
    The NALs contain different data, based on their types.
    """

    def __init__(self):
        self.nals: typing.List[NAL] = []

    def unpack(self, buf: bytes) -> bool:
        """
        Split the buffer into multiple NALs and store as a H264 object

        :param buf: The buffer to unpack into a H264 object
        :type buf: str
        :rtype: bool
        """
        nal_hdr = struct.pack(">L", NAL_HEADER)
        offsets = string_matching_boyer_moore_horspool(buf.decode(), nal_hdr.decode())

        for idx, offset in enumerate(offsets):
            if idx == len(offsets) - 1:
                nal_buf = buf[offset:]
            else:
                nal_buf = buf[offset : (offsets[idx + 1])]
            nal = NAL()
            nal.unpack(nal_buf)
            nal.offset = offset
            self.nals.append(nal)

        return True


class NAL(object):
    """
    The NAL can be split into the various types of NALs.
    """

    def __init__(self):
        self.type: int = 0
        self.size: int = 0
        self.sei: typing.Optional[STANAG4609_SEI] = None
        self.offset = 0

    def unpack(self, buf: bytes):
        """
        Split the buffer into a NAL object

        :param buf: The buffer to unpack into an NAL
        :type buf: str|bytes
        :rtype: bool
        """

        # First 4 bytes are the NAL_HEADER, then forbidden + type
        (self.type,) = struct.unpack_from(">B", buf, NAL_HEADER_LEN)
        self.type = self.type & 0x1F
        self.size = len(buf)
        if self.type == NAL_TYPES["SEI"]:
            sei = STANAG4609_SEI()
            sei.unpack(buf[(NAL_HEADER_LEN + 1) :])
            self.sei = sei

    def __len__(self):
        return self.size


def string_matching_boyer_moore_horspool(text="", pattern=""):
    """
    Returns positions where pattern is found in text.
    O(n)
    Performance: ord() is slow so we shouldn't use it here
    Example: text = 'ababbababa', pattern = 'aba'
         string_matching_boyer_moore_horspool(text, pattern) returns [0, 5, 7]
    @param text text to search inside
    @param pattern string to search for
    @return list containing offsets (shifts) where pattern is found inside text
    """
    m = len(pattern)
    n = len(text)
    offsets = []
    if m > n:
        return offsets
    skip = []
    for k in range(256):
        skip.append(m)
    for k in range(m - 1):
        my = pattern[k]
        if PY3:
            skip[pattern[k]] = m - k - 1
        else:
            skip[ord(pattern[k])] = m - k - 1
    skip = tuple(skip)
    k = m - 1
    while k < n:
        j = m - 1
        i = k
        while j >= 0 and text[i] == pattern[j]:
            j -= 1
            i -= 1
        if j == -1:
            offsets.append(i + 1)
        if PY3:
            k += skip[text[k]]
        else:
            k += skip[ord(text[k])]

    return offsets



import struct
from AcraNetwork import KMP
from AcraNetwork.IRIG106.Chapter11 import TS_CH4, TS_IEEE1558, PTPTime, RTCTime
import logging
import typing

DFLT_SYNC_WORD = 0xFE6B_2840

logger = logging.getLogger(__name__)
# logger.setLevel(logging.INFO)


class PCMMinorFrame(object):
    """
    Object that represents the PCM minor frame in a PCMPayload.
    """

    TS_LEN = 8
    ALIGN_16b = 0
    ALIGN_32b = 1
    DATA_HEADER_LEN = {ALIGN_16b: 2, ALIGN_32b: 4}
    DATA_HEADER_FORMAT = {ALIGN_16b: "<H", ALIGN_32b: "<I"}

    def __init__(
        self, ipts_source: typing.Optional[int] = TS_CH4, throughput: bool = False, alignment: int = ALIGN_16b
    ):
        if throughput:
            self.ipts = None
        elif ipts_source == TS_CH4:
            self.ipts = RTCTime()
        else:
            self.ipts = PTPTime()
        self._ipts_source = ipts_source
        self.throughput = throughput
        self.intra_packet_data_header = None
        self.minor_frame_data = bytes()
        self.alignment = alignment
        self.syncword = None
        self.sfid = None

    @property
    def payload(self) -> bytes:
        return self.minor_frame_data

    def unpack(self, buffer, extract_sync_sfid=False):
        """
        Convert a string buffer into a PCMDataPacket
        :type buffer: str
        :rtype: bool
        """
        if self.ipts is not None:
            self.ipts.unpack(buffer[:8])
        if not self.throughput:
            (self.intra_packet_data_header,) = struct.unpack_from(
                PCMMinorFrame.DATA_HEADER_FORMAT[self.alignment], buffer, 8
            )
            if extract_sync_sfid:
                (msw, lsw, self.sfid) = struct.unpack_from(
                    "<HHH", buffer, 8 + PCMMinorFrame.DATA_HEADER_LEN[self.alignment]
                )
                self.syncword = lsw + (msw << 16)
            self.minor_frame_data = buffer[PCMMinorFrame.DATA_HEADER_LEN[self.alignment] + PCMMinorFrame.TS_LEN :]
        else:
            self.minor_frame_data = buffer
        return True

    def pack(self):
        """
        Convert a PCMFrame object into a string buffer
        :return:
        """
        if self.throughput:
            buf = self.minor_frame_data
        else:
            if self.ipts is None:
                raise Exception("Timestamp should be defined in non-throughput mode")
            buf = self.ipts.pack() + struct.pack(
                PCMMinorFrame.DATA_HEADER_FORMAT[self.alignment], self.intra_packet_data_header
            )
            if self.syncword is not None:
                buf += struct.pack(">I", self.syncword)
            if self.sfid is not None:
                buf += struct.pack(">H", self.sfid)
            buf += self.minor_frame_data

        return buf

    def __repr__(self):
        if self.throughput:
            _str = f"Minor Frame Throughput mode Time={self.ipts} Payload_len={len(self.minor_frame_data)}"
        else:
            _str = "Minor Frame. Time={} DataHdr={:#0X} Payload_len={}".format(
                self.ipts, self.intra_packet_data_header, len(self.minor_frame_data)
            )
        return _str

    def __eq__(self, other):
        if not isinstance(other, PCMMinorFrame):
            return False
        for attr in ["ipts", "intra_packet_data_header", "minor_frame_data", "syncword", "sfid", "throughput"]:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    def __ne__(self, other):
        return not self.__eq__(other)


PCM_DATA_FRAME_FILL = 0x0
MODE_THROUGHPUT = 0x1 << 20
MODE_ALIGNMENT = 0x1 << 21


class PCMDataPacket(object):
    """
    This object represents the Payload to a Chapter 10 PCM packet

    The user needs to tell the object how many minor frames in Payload before unpacking a buffer.
    :type minor_frames: [PCMMinorFrame]


    >>> from base64 import b64decode
    >>> buffer = b64decode('AAAQAGv+QCg=')
    >>> p = PCMDataPacket(syncword=DFLT_SYNC_WORD)
    >>> p.unpack(buffer)
    True
    >>> print(p)
    PCM Data Packet Format 1. Channel Specific Word =0X100000
    Minor Frame Throughput mode Time=None Payload_len=4
    <BLANKLINE>
    """

    def __init__(
        self,
        ipts_source: typing.Optional[int] = TS_CH4,
        syncword: typing.Optional[int] = None,
        minor_frame_size_bytes: typing.Optional[int] = None,
    ):
        self.channel_specific_word: int = 0
        self._ipts_source: typing.Optional[int] = ipts_source
        self.minor_frame_size_bytes: typing.Optional[int] = minor_frame_size_bytes
        self.syncword = syncword
        self.minor_frames: typing.List[PCMMinorFrame] = []

    def unpack(self, buffer: bytes, extract_sync_sfid: bool = False) -> bool:
        """
        Convert a string buffer into a PCMDataPacket
        :type buffer: bytes
        :rtype: bool
        """

        (self.channel_specific_word,) = struct.unpack_from("<I", buffer, 0)
        channel_specific_len = 4
        throughput_mode = bool(self.channel_specific_word & MODE_THROUGHPUT == MODE_THROUGHPUT)
        alignment_mode = int(self.channel_specific_word & MODE_ALIGNMENT == MODE_ALIGNMENT)
        if throughput_mode:
            minor_frame = PCMMinorFrame(throughput=True, alignment=alignment_mode)
            minor_frame.unpack(buffer[channel_specific_len:])
            self.minor_frames.append(minor_frame)
        else:
            offset = channel_specific_len
            if self.minor_frame_size_bytes is None and self.syncword is None:
                self.minor_frame_size_bytes = (
                    len(buffer)
                    - PCMMinorFrame.TS_LEN
                    - PCMMinorFrame.DATA_HEADER_LEN[alignment_mode]
                    - channel_specific_len
                )
            elif self.minor_frame_size_bytes is None:
                # Search the payload for some sync words
                kmp = KMP()
                sync_word_offsets = kmp.search(buffer, struct.pack(">I", self.syncword))
                # More than one sync word. So split the payload
                if len(sync_word_offsets) < 2:
                    self.minor_frame_size_bytes = (
                        len(buffer)
                        - PCMMinorFrame.TS_LEN
                        - PCMMinorFrame.DATA_HEADER_LEN[alignment_mode]
                        - channel_specific_len
                    )
                else:
                    self.minor_frame_size_bytes = (
                        sync_word_offsets[1]
                        - sync_word_offsets[0]
                        - PCMMinorFrame.TS_LEN
                        - PCMMinorFrame.DATA_HEADER_LEN[alignment_mode]
                    )
                    logger.debug(
                        f"Found multiple sync words. Getting minor frame size as {self.minor_frame_size_bytes}"
                    )

            _byte_count_req = (
                self.minor_frame_size_bytes + PCMMinorFrame.TS_LEN + PCMMinorFrame.DATA_HEADER_LEN[alignment_mode]
            )
            while offset + _byte_count_req <= len(buffer):
                minor_frame = PCMMinorFrame(self._ipts_source, alignment=alignment_mode)
                if (_byte_count_req) % 2 != 0:
                    padding = 1
                else:
                    padding = 0
                try:
                    minor_frame.unpack(buffer[offset : offset + _byte_count_req], extract_sync_sfid=extract_sync_sfid)
                except Exception as e:
                    raise Exception(
                        "Unpacking payload at offset {} of {} failed. Err={}".format(offset, len(buffer), e)
                    )
                offset += _byte_count_req + padding
                self.minor_frames.append(minor_frame)

        return True

    def pack(self):
        buf = struct.pack("<I", self.channel_specific_word)
        throughput_mode = bool(self.channel_specific_word & MODE_THROUGHPUT == MODE_THROUGHPUT)
        for mf in self.minor_frames:
            buf += mf.pack()
            if len(mf.pack()) % 2 == 1:
                buf += struct.pack(">B", PCM_DATA_FRAME_FILL)

        return buf

    def append(self, minorframe):
        self.minor_frames.append(minorframe)

    def __repr__(self):
        _rstr = "PCM Data Packet Format 1. Channel Specific Word ={:#0X}\n".format(self.channel_specific_word)
        for m in self.minor_frames:
            _rstr += "{}\n".format(repr(m))

        return _rstr

    def __iter__(self):
        self._index = 0
        return self

    def next(self):
        if self._index < len(self.minor_frames):
            _frame = self.minor_frames[self._index]
            self._index += 1
            return _frame
        else:
            raise StopIteration

    __next__ = next

    def __getitem__(self, key):
        return self.minor_frames[key]

    def __eq__(self, other):
        """

        :type other: PCMDataPacket
        :return:
        """
        if not isinstance(other, PCMDataPacket):
            return False

        if self.channel_specific_word != other.channel_specific_word:
            return False

        if len(self.minor_frames) != len(other.minor_frames):
            return False

        for idx in range(len(self.minor_frames)):
            if self.minor_frames[idx] != other.minor_frames[idx]:
                return False

        return True
